import tempfile
import pkgutil
import setuptools

import rail
from rail.core import RailEnv
import rail.stages


def test_print_rail_packages():
    RailEnv.print_rail_packages()

    
def test_print_rail_namespaces():
    RailEnv.print_rail_namespaces()


def test_print_rail_modules():
    RailEnv.print_rail_modules()


def test_print_rail_namespace_tree():
    RailEnv.print_rail_namespace_tree()


def test_import_and_attach_all():
    rail.stages.import_and_attach_all()
    RailEnv.print_rail_stage_dict()


def test_api_rst():
    with tempfile.TemporaryDirectory() as tmpdirname:
        RailEnv.do_api_rst(tmpdirname)



"""
" Abstract base class defining an Evaluator

The key feature is that the evaluate method.
"""

import numpy as np

from ceci.config import StageParameter as Param
from rail.core.data import Hdf5Handle, QPHandle
from rail.core.stage import RailStage

from rail.evaluation.utils import stat_and_pval
from rail.evaluation.metrics.cdeloss import CDELoss
from qp.metrics.pit import PIT
from rail.evaluation.metrics.pointestimates import PointSigmaIQR, PointBias, PointOutlierRate, PointSigmaMAD


class Evaluator(RailStage):
    """Evaluate the performance of a photo-Z estimator """

    name = 'Evaluator'
    config_options = RailStage.config_options.copy()
    config_options.update(zmin=Param(float, 0., msg="min z for grid"),
                          zmax=Param(float, 3.0, msg="max z for grid"),
                          nzbins=Param(int, 301, msg="# of bins in zgrid"),
                          pit_metrics=Param(str, 'all', msg='PIT-based metrics to include'),
                          point_metrics=Param(str, 'all', msg='Point-estimate metrics to include'),
                          do_cde=Param(bool, True, msg='Evaluate CDE Metric'))
    inputs = [('input', QPHandle),
              ('truth', Hdf5Handle)]
    outputs = [('output', Hdf5Handle)]

    def __init__(self, args, comm=None):
        """Initialize Evaluator"""
        RailStage.__init__(self, args, comm=comm)

    def evaluate(self, data, truth):
        """Evaluate the performance of an estimator

        This will attach the input data and truth to this `Evaluator`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Parameters
        ----------
        data : qp.Ensemble
            The sample to evaluate
        truth : Table-like
            Table with the truth information

        Returns
        -------
        output : Table-like
            The evaluation metrics
        """

        self.set_data('input', data)
        self.set_data('truth', truth)
        self.run()
        self.finalize()
        return self.get_handle('output')

    def run(self):
        """ Run method

        Evaluate all the metrics and put them into a table

        Notes
        -----
        Get the input data from the data store under this stages 'input' tag
        Get the truth data from the data store under this stages 'truth' tag
        Puts the data into the data store under this stages 'output' tag
        """

        pz_data = self.get_data('input')
        z_true = self.get_data('truth')['redshift']
        zgrid = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins+1)

        # Create an instance of the PIT class
        pitobj = PIT(pz_data, z_true)

        # Build reference dictionary of the PIT meta-metrics from this PIT instance
        PIT_METRICS = dict(
            AD=getattr(pitobj, 'evaluate_PIT_anderson_ksamp'),
            CvM=getattr(pitobj, 'evaluate_PIT_CvM'),
            KS=getattr(pitobj, 'evaluate_PIT_KS'),
            OutRate=getattr(pitobj, 'evaluate_PIT_outlier_rate'),
        )

        # Parse the input configuration to determine which meta-metrics should be calculated
        if self.config.pit_metrics == 'all':
            pit_metrics = list(PIT_METRICS.keys())
        else:  #pragma: no cover
            pit_metrics = self.config.pit_metrics.split()

        # Evaluate each of the requested meta-metrics, and store the result in `out_table`
        out_table = {}
        for pit_metric in pit_metrics:
            value = PIT_METRICS[pit_metric]()

            # The result objects of some meta-metrics are bespoke scipy objects with inconsistent fields.
            # Here we do our best to store the relevant fields in `out_table`.
            if isinstance(value, list):  # pragma: no cover
                out_table[f'PIT_{pit_metric}'] = value
            else:
                out_table[f'PIT_{pit_metric}_stat'] = [getattr(value, 'statistic', None)]
                out_table[f'PIT_{pit_metric}_pval'] = [getattr(value, 'p_value', None)]
                out_table[f'PIT_{pit_metric}_significance_level'] = [getattr(value, 'significance_level', None)]

        POINT_METRICS = dict(SimgaIQR=PointSigmaIQR,
                             Bias=PointBias,
                             OutlierRate=PointOutlierRate,
                             SigmaMAD=PointSigmaMAD)
        if self.config.point_metrics == 'all':
            point_metrics = list(POINT_METRICS.keys())
        else:  #pragma: no cover
            point_metrics = self.config.point_metrics.split()

        z_mode = None
        for point_metric in point_metrics:
            if z_mode is None:
                z_mode = np.squeeze(pz_data.mode(grid=zgrid))
            value = POINT_METRICS[point_metric](z_mode, z_true).evaluate()
            out_table[f'POINT_{point_metric}'] = [value]

        if self.config.do_cde:
            value = CDELoss(pz_data, zgrid, z_true).evaluate()
            out_table['CDE_stat'] = [value.statistic]
            out_table['CDE_pval'] = [value.p_value]

        self.add_data('output', out_table)



""" Abstract base class defining a degrader

The key feature is that the __call__ method takes a pandas DataFrame and a
seed, and returns a pandas DataFrame, and wraps the run method
"""

from rail.core.stage import RailStage
from rail.core.data import PqHandle

class Degrader(RailStage):
    """Base class Degraders, which apply various degradations to synthetic photometric data

    Degraders take "input" data in the form of pandas dataframes in Parquet files and
    provide as "output" another pandas dataframes written to Parquet files
    """

    name = 'Degrader'
    config_options = RailStage.config_options.copy()
    config_options.update(seed=12345)
    inputs = [('input', PqHandle)]
    outputs = [('output', PqHandle)]

    def __init__(self, args, comm=None):
        """Initialize Degrader that can degrade photometric data"""
        RailStage.__init__(self, args, comm=comm)

    def __call__(self, sample, seed: int = None):
        """The main interface method for `Degrader`

        Applies degradation.

        This will attach the sample to this `Degrader`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Finally, this will return a PqHandle providing access to that output data.

        Parameters
        ----------
        sample : table-like
            The sample to be degraded
        seed : int, default=None
            An integer to set the numpy random seed

        Returns
        -------
        output_data : `PqHandle`
            A handle giving access to a table with degraded sample
        """
        if seed is not None:
            self.config.seed = seed
        self.set_data('input', sample)
        self.run()
        self.finalize()
        return self.get_handle('output')



""" Applying selection functions to catalog """

import os

import numpy as np
from ceci.config import StageParameter as Param
from rail.creation.degrader import Degrader
from scipy.interpolate import interp1d
from rail.core.utils import RAILDIR


class SpecSelection(Degrader):
    """
    The super class of spectroscopic selections.

    Parameters
    ----------
    N_tot: integer
        total number of down-sampled, spec-selected galaxies.
        If N_tot is greater than the number of spec-sepected galaxies, then
        it will be ignored.
    nondetect_val: value to be removed for non detects
    downsample: bool
        If True, then downsample the pre-selected galaxies
        to N_tot galaxies.
    success_rate_dir: string, the path to the success rate files.
    percentile_cut: If using color-based redshift cut, percentile in redshifts above which redshifts will be cut from the sample. Default is 100 (no cut)
    colnames: a dictionary that includes necessary columns
                         (magnitudes, colors and redshift) for selection. For magnitudes, the keys are ugrizy; for colors, the keys are,
                         for example, gr standing for g-r; for redshift, the key is 'redshift'.
    random_seed: random seed for reproducibility.

    """

    name = "specselection"
    config_options = Degrader.config_options.copy()
    config_options.update(
        N_tot=Param(int, 10000, msg="Number of selected sources"),
        nondetect_val=Param(float, 99.0, msg="value to be removed for non detects"),
        downsample=Param(
            bool,
            True,
            msg="If true, downsample the selected sources into a total number of N_tot",
        ),
        success_rate_dir=Param(
            str,
            os.path.join(RAILDIR,
                "rail/examples_data/creation_data/data/success_rate_data",
            ),
            msg="The path to the directory containing success rate files.",
        ),
        percentile_cut=Param(int, 100, msg="cut redshifts above this percentile"),
        colnames=Param(
            dict,
            {
                **{band: "mag_" + band + "_lsst" for band in "ugrizy"},
                **{"redshift": "redshift"},
            },
            msg="a dictionary that includes necessary columns\
                         (magnitudes, colors and redshift) for selection. For magnitudes, the keys are ugrizy; for colors, the keys are, \
                         for example, gr standing for g-r; for redshift, the key is 'redshift'",
        ),
        random_seed=Param(int, 42, msg="random seed for reproducibility"),
    )

    def __init__(self, args, comm=None):
        Degrader.__init__(self, args, comm=comm)
        # validate the settings
        self._validate_settings()
        self.mask = None
        self.rng = None

    def _validate_settings(self):
        """
        Validate all the settings.
        """

        if self.config.N_tot < 0:
            raise ValueError(
                "Total number of selected sources must be a " "positive integer."
            )
        if os.path.exists(self.config.success_rate_dir) is not True:
            raise ValueError(
                "Success rate path: "
                + self.config.success_rate_dir
                + " does not exist!"
            )

    def validate_colnames(self, data):
        """
        Validate the column names of data table to make sure they have necessary information
        for each selection.
        colnames: a list of column names.
        """
        colnames = self.config.colnames.values()
        check = all(item in data.columns for item in colnames)
        if check is not True:
            raise ValueError(
                "Columns in the data are not enough for the selection."
                + "The data should contain "
                + str(list(colnames))
                + ". \n"
            )

    def selection(self, data):
        """
        Selection functions. This should be overwritten by the subclasses
        corresponding to different spec selections.
        """

    def invalid_cut(self, data):
        """
        This function removes entries in the data that have invalid magnitude values
        (nondetect_val or NaN)
        """
        nondetect_val = self.config.nondetect_val
        for band in "ugrizy":
            if band not in self.config.colnames.keys():
                continue
            colname = self.config.colnames[band]
            self.mask &= (np.abs(data[colname]) < nondetect_val) & (
                ~np.isnan(data[colname])
            )

    def downsampling_N_tot(self):
        """
        Method to randomly sample down the objects to a given
        number of data objects.
        """
        N_tot = self.config.N_tot
        N_selected = np.count_nonzero(self.mask)
        if N_tot > N_selected:
            print(
                "Warning: N_tot is greater than the size of spec-selected "
                + "sample ("
                + str(N_selected)
                + "). The spec-selected sample "
                + "is returned."
            )
            return
        else:
            idx_selected = np.where(self.mask)[0]
            idx_keep = self.rng.choice(idx_selected, replace=False, size=N_tot)
            # create a mask with only those entries enabled that have been
            # selected
            mask = np.zeros_like(self.mask)
            mask[idx_keep] = True
            # update the internal state
            self.mask &= mask

    def run(self):
        """
        Run the selection
        """
        self.rng = np.random.default_rng(seed=self.config.seed)
        # get the bands and bandNames present in the data
        data = self.get_data("input", allow_missing=True)
        self.validate_colnames(data)
        self.mask = np.product(~np.isnan(data.to_numpy()), axis=1)
        self.invalid_cut(data)
        self.selection(data)
        if self.config.downsample is True:
            self.downsampling_N_tot()

        data_selected = data.iloc[np.where(self.mask == 1)[0]]

        self.add_data("output", data_selected)

    def __repr__(self):
        """
        Define how the model is represented and printed.
        """


class SpecSelection_GAMA(SpecSelection):
    """
    The class of spectroscopic selections with GAMA.
    The GAMA survey covers an area of 286 deg^2, with ~238000 objects
    The necessary column is r band
    """

    name = "specselection_gama"

    def selection(self, data):
        """
        GAMA selection function
        """
        print("Applying the selection from GAMA survey...")
        self.mask *= data[self.config.colnames["r"]] < 19.87

    def __repr__(self):
        """
        Define how the model is represented and printed.
        """
        # start message
        printMsg = "Applying the GAMA selection."

        return printMsg


class SpecSelection_BOSS(SpecSelection):
    """
    The class of spectroscopic selections with BOSS.
    BOSS selection function is based on
    http://www.sdss3.org/dr9/algorithms/boss_galaxy_ts.php
    The selection has changed slightly compared to Dawson+13
    BOSS covers an area of 9100 deg^2 with 893,319 galaxies.
    For BOSS selection, the data should at least include gri bands.
    """

    name = "specselection_boss"

    def selection(self, data):
        """
        The BOSS selection function.
        """

        print("Applying the selection from BOSS survey...")
        # cut quantities (unchanged)
        c_p = 0.7 * (
            data[self.config.colnames["g"]] - data[self.config.colnames["r"]]
        ) + 1.2 * (
            data[self.config.colnames["r"]] - data[self.config.colnames["i"]] - 0.18
        )
        c_r = (
            (data[self.config.colnames["r"]] - data[self.config.colnames["i"]])
            - (data[self.config.colnames["g"]] - data[self.config.colnames["r"]]) / 4.0
            - 0.18
        )
        d_r = (data[self.config.colnames["r"]] - data[self.config.colnames["i"]]) - (
            data[self.config.colnames["g"]] - data[self.config.colnames["r"]]
        ) / 8.0
        # defining the LOWZ sample
        # we cannot apply the r_psf - r_cmod cut
        low_z = (
            (data[self.config.colnames["r"]] > 16.0)
            & (data[self.config.colnames["r"]] < 20.0)
            & (np.abs(c_r) < 0.2)  # 19.6
            & (data[self.config.colnames["r"]] < 13.35 + c_p / 0.3)
        )  # 13.5, 0.3
        # defining the CMASS sample
        # we cannot apply the i_fib2, i_psf - i_mod and z_psf - z_mod cuts
        cmass = (
            (data[self.config.colnames["i"]] > 17.5)
            & (data[self.config.colnames["i"]] < 20.1)
            & (d_r > 0.55)  # 19.9
            & (data[self.config.colnames["i"]] < 19.98 + 1.6 * (d_r - 0.7))
            & (  # 19.86, 1.6, 0.8
                (data[self.config.colnames["r"]] - data[self.config.colnames["i"]])
                < 2.0
            )
        )
        # NOTE: we ignore the CMASS sparse sample
        self.mask *= low_z | cmass

    def __repr__(self):
        """
        Define how the model is represented and printed.
        """
        # start message
        printMsg = "Applying the BOSS selection."

        return printMsg


class SpecSelection_DEEP2(SpecSelection):
    """
    The class of spectroscopic selections with DEEP2.
    DEEP2 has a sky coverage of 2.8 deg^2 with ~53000 spectra.
    For DEEP2, one needs R band magnitude, B-R/R-I colors which are not available for the time being.
    So we use LSST gri bands now. When the conversion degrader is ready, this subclass will be updated
    accordingly.
    """

    name = "specselection_deep2"

    def photometryCut(self, data):
        """
        Applying DEEP2 photometric cut based on Newman+13.
        This modified selection gives the best match to the data n(z) with
        its cut at z~0.75 and the B-R/R-I distribution (Newman+13, Fig. 12)

        Notes
        -----
        We cannot apply the surface brightness cut and do not apply the
        Gaussian weighted sampling near the original colour cuts.
        """
        mask = (
            (data[self.config.colnames["r"]] > 18.5)
            & (data[self.config.colnames["r"]] < 24.1)
            & (  # 24.1
                (
                    data[self.config.colnames["g"]] - data[self.config.colnames["r"]]
                    < 2.45
                    * (
                        data[self.config.colnames["r"]]
                        - data[self.config.colnames["i"]]
                    )
                    - 0.2976
                )
                |
                # 2.45, 0.2976
                (
                    data[self.config.colnames["r"]] - data[self.config.colnames["i"]]
                    > 1.1
                )
                | (
                    data[self.config.colnames["g"]] - data[self.config.colnames["r"]]
                    < 0.5
                )
            )
        )  # 0.5
        # update the internal state
        self.mask &= mask

    def speczSuccess(self, data):
        """
        Spec-z success rate as function of r_AB for Q>=3 read of Figure 13 in
        Newman+13 for DEEP2 fields 2-4. Values are binned in steps of 0.2 mag
        with the first and last bin centered on 19 and 24.
        """
        success_R_bins = np.arange(18.9, 24.1 + 0.01, 0.2)
        success_R_centers = (success_R_bins[1:] + success_R_bins[:-1]) / 2.0
        # paper has given 1 - [sucess rate] in the histogram
        success_rate_dir = self.config.success_rate_dir
        success_R_rate = np.loadtxt(os.path.join(success_rate_dir, "DEEP2_success.txt"))
        # interpolate the success rate as probability of being selected with
        # the probability at R > 24.1 being 0
        p_success_R = interp1d(
            success_R_centers,
            success_R_rate,
            kind="quadratic",
            bounds_error=False,
            fill_value=(success_R_rate[0], 0.0),
        )
        # Randomly sample objects according to their success rate
        random_draw = self.rng.random(len(data))
        mask = random_draw < p_success_R(data[self.config.colnames["r"]])
        # update the internal state
        self.mask &= mask

    def selection(self, data):
        """
        DEEP2 selection function
        """
        self.photometryCut(data)
        self.speczSuccess(data)

    def __repr__(self):
        """
        Define how the model is represented and printed.
        """
        # start message
        printMsg = "Applying the DEEP2 selection."

        return printMsg


class SpecSelection_VVDSf02(SpecSelection):
    """
    The class of spectroscopic selections with VVDSf02.
    It covers an area of 0.5 deg^2 with ~10000 sources.
    Necessary columns are i band magnitude and redshift.
    """

    name = "specselection_VVDSf02"

    def photometryCut(self, data):
        """
        Photometric cut of VVDS 2h-field based on LeFèvre+05.

        Notes
        -----
        The oversight of 1.0 magnitudes on the bright end misses 0.2% of galaxies.
        """
        mask = (data[self.config.colnames["i"]] > 17.5) & (
            data[self.config.colnames["i"]] < 24.0
        )
        # 17.5, 24.0
        self.mask &= mask

    def speczSuccess(self, data):
        """
        Success rate of VVDS 2h-field

        Notes
        -----
        We use a redshift-based and I-band based success rate
        independently here since we do not know their correlation,
        which makes the success rate worse than in reality.

        Spec-z success rate as function of i_AB read of Figure 16 in
        LeFevre+05 for the VVDS 2h field. Values are binned in steps of
        0.5 mag with the first starting at 17 and the last bin ending at 24.
        """
        success_I_bins = np.arange(17.0, 24.0 + 0.01, 0.5)
        success_I_centers = (success_I_bins[1:] + success_I_bins[:-1]) / 2.0
        success_rate_dir = self.config.success_rate_dir
        success_I_rate = np.loadtxt(
            os.path.join(success_rate_dir, "VVDSf02_I_success.txt")
        )
        # interpolate the success rate as probability of being selected with
        # the probability at I > 24 being 0
        p_success_I = interp1d(
            success_I_centers,
            success_I_rate,
            kind="quadratic",
            bounds_error=False,
            fill_value=(success_I_rate[0], 0.0),
        )
        # Randomly sample objects according to their success rate
        random_draw = self.rng.random(len(data))
        mask = random_draw < p_success_I(data["mag_i_lsst"])
        # Spec-z success rate as function of redshift read of Figure 13a/b in
        # LeFevre+13 for VVDS deep sample. The listing is split by i_AB into
        # ranges (17.5; 22.5] and (22.5; 24.0].
        # NOTE: at z > 1.75 there are only lower limits (due to a lack of
        # spec-z?), thus the success rate is extrapolated as 1.0 at z > 1.75
        success_z_bright_centers, success_z_bright_rate = np.loadtxt(
            os.path.join(success_rate_dir, "VVDSf02_z_bright_success.txt")
        ).T
        success_z_deep_centers, success_z_deep_rate = np.loadtxt(
            os.path.join(success_rate_dir, "VVDSf02_z_deep_success.txt")
        ).T
        # interpolate the success rates as probability of being selected with
        # the probability in the bright bin at z > 1.75 being 1.0 and the deep
        # bin at z > 4.0 being 0.0
        p_success_z_bright = interp1d(
            success_z_bright_centers,
            success_z_bright_rate,
            kind="quadratic",
            bounds_error=False,
            fill_value=(success_z_bright_rate[0], 1.0),
        )
        p_success_z_deep = interp1d(
            success_z_deep_centers,
            success_z_deep_rate,
            kind="quadratic",
            bounds_error=False,
            fill_value=(success_z_deep_rate[0], 0.0),
        )
        # Randomly sample objects according to their success rate
        random_draw = self.rng.random(len(data))
        iterator = zip(
            [
                data[self.config.colnames["i"]] <= 22.5,
                data[self.config.colnames["i"]] > 22.5,
            ],
            [p_success_z_bright, p_success_z_deep],
        )
        for m, p_success_z in iterator:
            mask[m] &= random_draw[m] < p_success_z(
                data[self.config.colnames["redshift"]][m]
            )
        # update the internal state
        self.mask &= mask

    def selection(self, data):
        self.photometryCut(data)
        self.speczSuccess(data)

    def __repr__(self):
        """
        Define how the model is represented and printed.
        """
        # start message
        printMsg = "Applying the VVDSf02 selection."

        return printMsg


class SpecSelection_zCOSMOS(SpecSelection):
    """
    The class of spectroscopic selections with zCOSMOS
    It covers an area of 1.7 deg^2 with ~20000 galaxies.
    For zCOSMOS, the data should at least include i band and redshift.
    """

    name = "specselection_zCOSMOS"

    def photometryCut(self, data):
        """
        Photometry cut for zCOSMOS based on Lilly+09.
        NOTE: This only includes zCOSMOS bright.
        update the internal state
        """
        mask = (data[self.config.colnames["i"]] > 15.0) & (
            data[self.config.colnames["i"]] < 22.5
        )
        # 15.0, 22.5
        self.mask &= mask

    def speczSuccess(self, data):
        """
        Spec-z success rate as function of redshift (x) and I_AB (y) read of
        Figure 3 in Lilly+09 for zCOSMOS bright sample.
        """
        success_rate_dir = self.config.success_rate_dir
        x = np.arange(0, 1.4, 0.00587002, dtype=np.float64)
        y = np.arange(18, 22.4, 0.01464226, dtype=np.float64)

        pixels_y = np.searchsorted(y, data[self.config.colnames["i"]])
        pixels_x = np.searchsorted(x, data[self.config.colnames["redshift"]])

        rates = np.loadtxt(os.path.join(success_rate_dir, "zCOSMOS_success.txt"))
        ratio_list = np.zeros(len(pixels_y))
        for i, py in enumerate(pixels_y):
            if (
                (py >= rates.shape[0])
                or (pixels_x[i] >= rates.shape[1])
                or (py == 0)
                or (pixels_x[i] == 0)
            ):
                ratio_list[i] = 0
            else:
                ratio_list[i] = rates[pixels_y[i] - 1][pixels_x[i] - 1]

        randoms = self.rng.uniform(size=data[self.config.colnames["i"]].size)
        mask = randoms <= ratio_list
        self.mask &= mask

    def selection(self, data):
        self.photometryCut(data)
        self.speczSuccess(data)

    def __repr__(self):
        """
        Define how the model is represented and printed.
        """
        # start message
        printMsg = "Applying the zCOSMOS selection."

        return printMsg


class SpecSelection_HSC(SpecSelection):
    """
    The class of spectroscopic selections with HSC
    or HSC, the data should at least include giz bands and redshift.
    """

    name = "specselection_HSC"

    def photometryCut(self, data):
        """
        HSC galaxies were binned in color magnitude space with i-band mag from -2 to 6 and g-z color from 13 to 26.
        """
        mask = (data[self.config.colnames["i"]] > 13.0) & (
            data[self.config.colnames["i"]] < 26.0
        )
        self.mask &= mask
        gz = data[self.config.colnames["g"]] - data[self.config.colnames["z"]]
        mask = (gz > -2.0) & (gz < 6.0)
        self.mask &= mask

    def speczSuccess(self, data):
        """
        HSC galaxies were binned in color magnitude space with i-band mag from -2 to 6 and g-z color from 13 to 26
        200 bins in each direction. The ratio of of galaxies with spectroscopic redshifts (training galaxies) to
        galaxies with only photometry in HSC wide field (application galaxies) was computed for each pixel. We divide
        the data into the same pixels and randomly select galaxies into the training sample based on the HSC ratios
        """
        success_rate_dir = self.config.success_rate_dir
        x_edge = np.linspace(13, 26, 201, endpoint=True)
        y_edge = np.linspace(-2, 6, 201, endpoint=True)

        rates = np.loadtxt(os.path.join(success_rate_dir, "hsc_success.txt"))

        pixels_y = np.searchsorted(
            y_edge, data[self.config.colnames["g"]] - data[self.config.colnames["z"]]
        )
        pixels_x = np.searchsorted(x_edge, data[self.config.colnames["i"]])

        # Do the color-based, percentile-based redshift cut

        percentile_cut = self.config.percentile_cut

        mask_keep = np.ones_like(data[self.config.colnames["i"]])
        if percentile_cut != 100:  # pragma: no cover
            pixels_y_unique = np.unique(pixels_y)
            pixels_x_unique = np.unique(pixels_x)

            for y in pixels_y_unique:
                for x in pixels_x_unique:
                    ind_inpix = np.where((pixels_y == y) * (pixels_x == x))[0]
                    if ind_inpix.size == 0:
                        continue
                    redshifts = data[self.config.colnames["redshift"]][ind_inpix]
                    percentile = np.percentile(redshifts, percentile_cut)
                    ind_remove = ind_inpix[redshifts > percentile]
                    mask_keep[ind_remove] = 0
            self.mask &= mask_keep

        pixels_y = pixels_y - 1
        pixels_x = pixels_x - 1

        ratio_list = np.zeros(len(pixels_y))
        for i, py in enumerate(pixels_y):
            if (py >= rates.shape[0]) or (pixels_x[i] >= rates.shape[1]):
                ratio_list[i] = 0
            else:
                ratio_list[i] = rates[pixels_y[i]][pixels_x[i]]

        randoms = self.rng.uniform(size=data[self.config.colnames["i"]].size)
        mask = randoms <= ratio_list
        self.mask &= mask

    def selection(self, data):
        self.photometryCut(data)
        self.speczSuccess(data)

    def __repr__(self):
        """
        Define how the model is represented and printed.
        """
        # start message
        printMsg = "Applying the HSC selection."

        return printMsg



"""Core code for RAIL"""
import pkgutil
import setuptools
import rail
import os

def find_version():
    """Find the version"""
    # setuptools_scm should install a
    # file _version alongside this one.
    from . import _version
    return _version.version

try:
    __version__ = find_version()
except ImportError: # pragma: no cover
    __version__ = "unknown"

from .stage import RailPipeline, RailStage
#from .utilPhotometry import PhotormetryManipulator, HyperbolicSmoothing, HyperbolicMagnitudes
from .utilStages import ColumnMapper, RowSelector, TableConverter
from .introspection import RailEnv



import os

import numpy as np
import qp

import rail.evaluation.metrics.pointestimates as pe
from rail.core.data import QPHandle, TableHandle
from rail.core.stage import RailStage
from rail.evaluation.evaluator import Evaluator
from rail.evaluation.metrics.cdeloss import CDELoss

# values for metrics
OUTRATE = 0.0
KSVAL = 0.367384
CVMVAL = 20.63155
ADVAL_ALL = 82.51480
ADVAL_CUT = 1.10750
CDEVAL = -4.31200
SIGIQR = 0.0045947
BIAS = -0.00001576
OUTRATE = 0.0
SIGMAD = 0.0046489


def construct_test_ensemble():
    np.random.seed(87)
    nmax = 2.5
    NPDF = 399
    true_zs = np.random.uniform(high=nmax, size=NPDF)
    locs = np.expand_dims(true_zs + np.random.normal(0.0, 0.01, NPDF), -1)
    true_ez = (locs.flatten() - true_zs) / (1.0 + true_zs)
    scales = np.ones((NPDF, 1)) * 0.1 + np.random.uniform(size=(NPDF, 1)) * 0.05
    n_ens = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))
    zgrid = np.linspace(0, nmax, 301)
    grid_ens = n_ens.convert_to(qp.interp_gen, xvals=zgrid)
    return zgrid, true_zs, grid_ens, true_ez


def test_point_metrics():
    zgrid, zspec, pdf_ens, true_ez = construct_test_ensemble()
    zb = pdf_ens.mode(grid=zgrid).flatten()

    ez = pe.PointStatsEz(zb, zspec).evaluate()
    assert np.allclose(ez, true_ez, atol=1.0e-2)
    # grid limits ez vals to ~10^-2 tol

    sig_iqr = pe.PointSigmaIQR(zb, zspec).evaluate()
    assert np.isclose(sig_iqr, SIGIQR)

    bias = pe.PointBias(zb, zspec).evaluate()
    assert np.isclose(bias, BIAS)

    out_rate = pe.PointOutlierRate(zb, zspec).evaluate()
    assert np.isclose(out_rate, OUTRATE)

    sig_mad = pe.PointSigmaMAD(zb, zspec).evaluate()
    assert np.isclose(sig_mad, SIGMAD)


def test_evaluation_stage():
    DS = RailStage.data_store
    zgrid, zspec, pdf_ens, true_ez = construct_test_ensemble()
    pdf = DS.add_data("pdf", pdf_ens, QPHandle)
    truth_table = dict(redshift=zspec)
    truth = DS.add_data("truth", truth_table, TableHandle)
    evaluator = Evaluator.make_stage(name="Eval")
    evaluator.evaluate(pdf, truth)

    os.remove(evaluator.get_output(evaluator.get_aliased_tag("output"), final_name=True))



"""
A summarizer that simple makes a histogram of a point estimate
"""

import numpy as np
from ceci.config import StageParameter as Param
from rail.estimation.summarizer import PZSummarizer
from rail.core.data import QPHandle
import qp


class NaiveStackSummarizer(PZSummarizer):
    """Summarizer which simply histograms a point estimate
    """

    name = 'NaiveStackSummarizer'
    config_options = PZSummarizer.config_options.copy()
    config_options.update(zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
                          zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
                          nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
                          seed=Param(int, 87, msg="random seed"),
                          nsamples=Param(int, 1000, msg="Number of sample distributions to create"))
    outputs = [('output', QPHandle),
               ('single_NZ', QPHandle)]

    def __init__(self, args, comm=None):
        PZSummarizer.__init__(self, args, comm=comm)
        self.zgrid = None

    def run(self):
        rng = np.random.default_rng(seed=self.config.seed)
        test_data = self.get_data('input')
        self.zgrid = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins + 1)
        pdf_vals = test_data.pdf(self.zgrid)
        yvals = np.expand_dims(np.sum(np.where(np.isfinite(pdf_vals), pdf_vals, 0.), axis=0), 0)
        qp_d = qp.Ensemble(qp.interp, data=dict(xvals=self.zgrid, yvals=yvals))

        bvals = np.empty((self.config.nsamples, len(self.zgrid)))
        for i in range(self.config.nsamples):
            bootstrap_draws = rng.integers(low=0, high=test_data.npdf, size=test_data.npdf)
            bvals[i] = np.sum(pdf_vals[bootstrap_draws], axis=0)
        sample_ens = qp.Ensemble(qp.interp, data=dict(xvals=self.zgrid, yvals=bvals))

        self.add_data('output', sample_ens)
        self.add_data('single_NZ', qp_d)



"""Rail-specific data management"""

import os
import tables_io
import pickle
import qp



class DataHandle:
    """Class to act as a handle for a bit of data.  Associating it with a file and
    providing tools to read & write it to that file

    Parameters
    ----------
    tag : str
        The tag under which this data handle can be found in the store
    data : any or None
        The associated data
    path : str or None
        The path to the associated file
    creator : str or None
        The name of the stage that created this data handle
    """
    suffix = ''

    def __init__(self, tag, data=None, path=None, creator=None):
        """Constructor """
        self.tag = tag
        self.data = data
        self.path = path
        self.creator = creator
        self.fileObj = None
        self.groups = None
        self.partial = False

    def open(self, **kwargs):
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        if self.path is None:
            raise ValueError("DataHandle.open() called but path has not been specified")
        self.fileObj = self._open(os.path.expandvars(self.path), **kwargs)
        return self.fileObj

    @classmethod
    def _open(cls, path, **kwargs):
        raise NotImplementedError("DataHandle._open")  #pragma: no cover

    def close(self, **kwargs):  #pylint: disable=unused-argument
        """Close """
        self.fileObj = None

    def read(self, force=False, **kwargs):
        """Read and return the data from the associated file """
        if self.data is not None and not force:
            return self.data
        self.set_data(self._read(os.path.expandvars(self.path), **kwargs))
        return self.data

    def __call__(self, **kwargs):
        """Return the data, re-reading the fill if needed"""
        if self.has_data and not self.partial:
            return self.data
        return self.read(force=True, **kwargs)

    @classmethod
    def _read(cls, path, **kwargs):
        raise NotImplementedError("DataHandle._read")  #pragma: no cover

    def write(self, **kwargs):
        """Write the data to the associatied file """
        if self.path is None:
            raise ValueError("TableHandle.write() called but path has not been specified")
        if self.data is None:
            raise ValueError(f"TableHandle.write() called for path {self.path} with no data")
        outdir = os.path.dirname(os.path.abspath(os.path.expandvars(self.path)))
        if not os.path.exists(outdir):  #pragma: no cover
            os.makedirs(outdir, exist_ok=True)
        return self._write(self.data, os.path.expandvars(self.path), **kwargs)

    @classmethod
    def _write(cls, data, path, **kwargs):
        raise NotImplementedError("DataHandle._write")  #pragma: no cover

    def initialize_write(self, data_lenght, **kwargs):
        """Initialize file to be written by chunks"""
        if self.path is None:  #pragma: no cover
            raise ValueError("TableHandle.write() called but path has not been specified")
        self.groups, self.fileObj = self._initialize_write(self.data, os.path.expandvars(self.path), data_lenght, **kwargs)

    @classmethod
    def _initialize_write(cls, data, path, data_lenght, **kwargs):
        raise NotImplementedError("DataHandle._initialize_write") #pragma: no cover

    def write_chunk(self, start, end, **kwargs):
        """Write the data to the associatied file """
        if self.data is None:
            raise ValueError(f"TableHandle.write_chunk() called for path {self.path} with no data")
        if self.fileObj is None:
            raise ValueError(f"TableHandle.write_chunk() called before open for {self.tag} : {self.path}")
        return self._write_chunk(self.data, self.fileObj, self.groups, start, end, **kwargs)


    @classmethod
    def _write_chunk(cls, data, fileObj, groups, start, end, **kwargs):
        raise NotImplementedError("DataHandle._write_chunk")  #pragma: no cover

    def finalize_write(self, **kwargs):
        """Finalize and close file written by chunks"""
        if self.fileObj is None:  #pragma: no cover
            raise ValueError(f"TableHandle.finalize_wite() called before open for {self.tag} : {self.path}")
        self._finalize_write(self.data, self.fileObj, **kwargs)

    @classmethod
    def _finalize_write(cls, data, fileObj, **kwargs):
        raise NotImplementedError("DataHandle._finalize_write")  #pragma: no cover

    def iterator(self, **kwargs):
        """Iterator over the data"""
        #if self.data is not None:
        #    for i in range(1):
        #        yield i, -1, self.data
        return self._iterator(self.path, **kwargs)

    def set_data(self, data, partial=False):
        """Set the data for a chunk, and set the partial flag to true"""
        self.data = data
        self.partial = partial

    def size(self, **kwargs):
        """Return the size of the data associated to this handle"""
        return self._size(self.path, **kwargs)

    @classmethod
    def _size(cls, path, **kwargs):
        raise NotImplementedError("DataHandle._size")  #pragma: no cover

    @classmethod
    def _iterator(cls, path, **kwargs):
        raise NotImplementedError("DataHandle._iterator")  #pragma: no cover

    @property
    def has_data(self):
        """Return true if the data for this handle are loaded """
        return self.data is not None

    @property
    def has_path(self):
        """Return true if the path for the associated file is defined """
        return self.path is not None

    @property
    def is_written(self):
        """Return true if the associated file has been written """
        if self.path is None:
            return False
        return os.path.exists(os.path.expandvars(self.path))

    def __str__(self):
        s = f"{type(self)} "
        if self.has_path:
            s += f"{self.path}, ("
        else:
            s += "None, ("
        if self.is_written:
            s += "w"
        if self.has_data:
            s += "d"
        s += ")"
        return s

    @classmethod
    def make_name(cls, tag):
        """Construct and return file name for a particular data tag """
        if cls.suffix:
            return f"{tag}.{cls.suffix}"
        else:
            return tag  #pragma: no cover


class TableHandle(DataHandle):
    """DataHandle for single tables of data
    """
    suffix = None

    @classmethod
    def _open(cls, path, **kwargs):
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        return tables_io.io.io_open(path, **kwargs)  #pylint: disable=no-member

    @classmethod
    def _read(cls, path, **kwargs):
        """Read and return the data from the associated file """
        return tables_io.read(path, **kwargs)

    @classmethod
    def _write(cls, data, path, **kwargs):
        """Write the data to the associatied file """
        return tables_io.write(data, path, **kwargs)

    @classmethod
    def _size(cls, path, **kwargs):
        return tables_io.io.getInputDataLengthHdf5(path, **kwargs)

    @classmethod
    def _iterator(cls, path, **kwargs):
        """Iterate over the data"""
        return tables_io.iteratorNative(path, **kwargs)


class Hdf5Handle(TableHandle):  # pragma: no cover
    """DataHandle for a table written to HDF5"""
    suffix = 'hdf5'

    @classmethod
    def _initialize_write(cls, data, path, data_lenght, **kwargs):
        initial_dict = cls._get_allocation_kwds(data, data_lenght)
        comm = kwargs.get('communicator', None)
        group, fout = tables_io.io.initializeHdf5WriteSingle(path, groupname=None, comm=comm, **initial_dict)
        return group, fout

    @classmethod
    def _get_allocation_kwds(cls, data, data_lenght):
        keywords = {}
        for key, array in data.items():
            shape = list(array.shape)
            shape[0] = data_lenght
            keywords[key] = (shape, array.dtype)
        return keywords

    @classmethod
    def _write_chunk(cls, data, fileObj, groups, start, end, **kwargs):
        tables_io.io.writeDictToHdf5ChunkSingle(fileObj, data, start, end, **kwargs)

    @classmethod
    def _finalize_write(cls, data, fileObj, **kwargs):
        return tables_io.io.finalizeHdf5Write(fileObj, **kwargs)

class FitsHandle(TableHandle):
    """DataHandle for a table written to fits"""
    suffix = 'fits'


class PqHandle(TableHandle):
    """DataHandle for a parquet table"""
    suffix = 'pq'


class QPHandle(DataHandle):
    """DataHandle for qp ensembles
    """
    suffix = 'hdf5'

    @classmethod
    def _open(cls, path, **kwargs):
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        return tables_io.io.io_open(path, **kwargs)  #pylint: disable=no-member

    @classmethod
    def _read(cls, path, **kwargs):
        """Read and return the data from the associated file """
        return qp.read(path)

    @classmethod
    def _write(cls, data, path, **kwargs):
        """Write the data to the associatied file """
        return data.write_to(path)

    @classmethod
    def _initialize_write(cls, data, path, data_lenght, **kwargs):
        comm = kwargs.get('communicator', None)
        return data.initializeHdf5Write(path, data_lenght, comm)

    @classmethod
    def _write_chunk(cls, data, fileObj, groups, start, end, **kwargs):
        return data.writeHdf5Chunk(fileObj, start, end)

    @classmethod
    def _finalize_write(cls, data, fileObj, **kwargs):
        return data.finalizeHdf5Write(fileObj)


def default_model_read(modelfile):
    """Default function to read model files, simply used pickle.load"""
    return pickle.load(open(modelfile, 'rb'))


def default_model_write(model, path):
    """Write the model, this default implementation uses pickle"""
    with open(path, 'wb') as fout:
        pickle.dump(obj=model, file=fout, protocol=pickle.HIGHEST_PROTOCOL)


class ModelDict(dict):
    """
    A specialized dict to keep track of individual estimation models objects: this is just a dict these additional features

    1. Keys are paths
    2. There is a read(path, force=False) method that reads a model object and inserts it into the dictionary
    3. There is a single static instance of this class
    """
    def open(self, path, mode, **kwargs):  #pylint: disable=no-self-use
        """Open the file and return the file handle"""
        return open(path, mode, **kwargs)

    def read(self, path, force=False, reader=None, **kwargs):  #pylint: disable=unused-argument
        """Read a model into this dict"""
        if reader is None:
            reader = default_model_read
        if force or path not in self:
            model = reader(path)
            self.__setitem__(path, model)
            return model
        return self[path]

    def write(self, model, path, force=False, writer=None, **kwargs):  #pylint: disable=unused-argument
        """Write the model, this default implementation uses pickle"""
        if writer is None:
            writer = default_model_write
        if force or path not in self:
            self.__setitem__(path, model)
            writer(model, path)



class ModelHandle(DataHandle):
    """DataHandle for machine learning models
    """
    suffix = 'pkl'

    model_factory = ModelDict()

    @classmethod
    def _open(cls, path, **kwargs):
        """Open and return the associated file
        """
        kwcopy = kwargs.copy()
        if kwcopy.pop('mode', 'r') == 'w':
            return cls.model_factory.open(path, mode='wb', **kwcopy)
        return cls.model_factory.read(path, **kwargs)

    @classmethod
    def _read(cls, path, **kwargs):
        """Read and return the data from the associated file """
        return cls.model_factory.read(path, **kwargs)

    @classmethod
    def _write(cls, data, path, **kwargs):
        """Write the data to the associatied file """
        return cls.model_factory.write(data, path, **kwargs)


class DataStore(dict):
    """Class to provide a transient data store

    This class:
    1) associates data products with keys
    2) provides functions to read and write the various data produces to associated files
    """
    allow_overwrite = False

    def __init__(self, **kwargs):
        """ Build from keywords

        Note
        ----
        All of the values must be data handles of this will raise a TypeError
        """
        dict.__init__(self)
        for key, val in kwargs.items():
            self[key] = val

    def __str__(self):
        """ Override __str__ casting to deal with `TableHandle` objects in the map """
        s = "{"
        for key, val in self.items():
            s += f"  {key}:{val}\n"
        s += "}"
        return s

    def __repr__(self):
        """ A custom representation """
        s = "DataStore\n"
        s += self.__str__()
        return s

    def __setitem__(self, key, value):
        """ Override the __setitem__ to work with `TableHandle` """
        if not isinstance(value, DataHandle):
            raise TypeError(f"Can only add objects of type DataHandle to DataStore, not {type(value)}")
        check = self.get(key)
        if check is not None and not self.allow_overwrite:
            raise ValueError(f"DataStore already has an item with key {key}, of type {type(check)}, created by {check.creator}")
        dict.__setitem__(self, key, value)
        return value

    def __getattr__(self, key):
        """ Allow attribute-like parameter access """
        try:
            return self.__getitem__(key)
        except KeyError as msg:
            # Kludge to get docstrings to work
            if key in ['__objclass__']:  #pragma: no cover
                return None
            raise KeyError from msg

    def __setattr__(self, key, value):
        """ Allow attribute-like parameter setting """
        return self.__setitem__(key, value)

    def add_data(self, key, data, handle_class, path=None, creator='DataStore'):
        """ Create a handle for some data, and insert it into the DataStore """
        handle = handle_class(key, path=path, data=data, creator=creator)
        self[key] = handle
        return handle

    def read_file(self, key, handle_class, path, creator='DataStore', **kwargs):
        """ Create a handle, use it to read a file, and insert it into the DataStore """
        handle = handle_class(key, path=path, data=None, creator=creator)
        handle.read(**kwargs)
        self[key] = handle
        return handle

    def read(self, key, force=False, **kwargs):
        """ Read the data associated to a particular key """
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to read data {key} because {msg}") from msg
        return handle.read(force, **kwargs)

    def open(self, key, mode='r', **kwargs):
        """ Open and return the file associated to a particular key """
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to open data {key} because {msg}") from msg
        return handle.open(mode=mode, **kwargs)

    def write(self, key, **kwargs):
        """ Write the data associated to a particular key """
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to write data {key} because {msg}") from msg
        return handle.write(**kwargs)

    def write_all(self, force=False, **kwargs):
        """ Write all the data in this DataStore """
        for key, handle in self.items():
            local_kwargs = kwargs.get(key, {})
            if handle.is_written and not force:
                continue
            handle.write(**local_kwargs)



_DATA_STORE = DataStore()

def DATA_STORE():
    """Return the factory instance"""
    return _DATA_STORE



import pkgutil
import setuptools
import os
import importlib

import rail


class RailEnv:

    PACKAGES = {}
    NAMESPACE_PATH_DICT = {}
    NAMESPACE_MODULE_DICT = {}
    MODULE_DICT = {}
    MODULE_PATH_DICT = {}
    TREE = {}
    STAGE_DICT = {}
    BASE_STAGES = []

    @classmethod
    def list_rail_packages(cls):
        """List all the packages that are available in the RAIL ecosystem"""
        cls.PACKAGES = {pkg.name:pkg for pkg in pkgutil.iter_modules(rail.__path__, rail.__name__ + '.')}
        return cls.PACKAGES

    
    @classmethod
    def print_rail_packages(cls):
        """Print all the packages that are available in the RAIL ecosystem"""
        if not cls.PACKAGES:
            cls.list_rail_packages()
        for pkg_name, pkg in cls.PACKAGES.items():
            print(f"{pkg_name} @ {pkg[0].path}")
        return

    @classmethod
    def list_rail_namespaces(cls):
        """List all the namespaces within rail"""
        cls.NAMESPACE_PATH_DICT.clear()
        
        for path_ in rail.__path__:
            namespaces = setuptools.find_namespace_packages(path_)
            for namespace_ in namespaces:
                # exclude stuff that starts with 'example'
                if namespace_.find('example') == 0:
                    continue
                if namespace_ in cls.NAMESPACE_PATH_DICT:  # pragma: no cover
                    cls.NAMESPACE_PATH_DICT[namespace_].append(path_)
                else:
                    cls.NAMESPACE_PATH_DICT[namespace_] = [path_]

        return cls.NAMESPACE_PATH_DICT

                    
    @classmethod
    def print_rail_namespaces(cls):
        """Print all the namespaces that are available in the RAIL ecosystem"""
        if not cls.NAMESPACE_PATH_DICT:
            cls.list_rail_namespaces()
        for key, val in cls.NAMESPACE_PATH_DICT.items():
            print(f"Namespace {key}")
            for vv in val:
                print(f"     {vv}")
        return


    @classmethod
    def list_rail_modules(cls):
        """List all modules within rail"""
        cls.MODULE_DICT.clear()
        cls.MODULE_PATH_DICT.clear()
        cls.NAMESPACE_MODULE_DICT.clear()
        if not cls.NAMESPACE_PATH_DICT:  # pragma: no cover
            cls.list_rail_namespaces()
        for key, val in cls.NAMESPACE_PATH_DICT.items():
            cls.NAMESPACE_MODULE_DICT[key] = []
            for vv in val:
                fullpath = os.path.join(vv, key.replace('.', '/'))
                modules = [pkg for pkg in pkgutil.iter_modules([fullpath], rail.__name__ + '.' + key + '.')]
                for module_ in modules:
                    if module_ in cls.MODULE_DICT:  # pragma: no cover
                        cls.MODULE_DICT[module_.name].append(key)
                    else:
                        cls.MODULE_DICT[module_.name] = [key]
                    cls.NAMESPACE_MODULE_DICT[key].append(module_)
                    cls.MODULE_PATH_DICT[module_.name] = module_[0].path

        return cls.MODULE_PATH_DICT


    @classmethod
    def print_rail_modules(cls):
        """Print all the moduels that are available in the RAIL ecosystem"""
        if not cls.MODULE_DICT:
            cls.list_rail_modules()
        
        for key, val in cls.MODULE_DICT.items():
            print(f"Module {key}")
            for vv in val:
                print(f"     {vv}")

        for key, val in cls.NAMESPACE_MODULE_DICT.items():
            print(f"Namespace {key}")
            for vv in val:
                print(f"     {vv}")
        return


    @classmethod
    def build_rail_namespace_tree(cls):
        """Build a tree of the namespaces and packages in rail"""
        cls.TREE.clear()
        if not cls.NAMESPACE_MODULE_DICT:  # pragma: no cover
            cls.list_rail_modules()

        if not cls.PACKAGES:  # pragma: no cover
            cls.list_rail_packages()

        level_dict = {}
        for key in cls.NAMESPACE_MODULE_DICT.keys():
            count = key.count('.')
            if count in level_dict:
                level_dict[count].append(key)
            else:
                level_dict[count] = [key]

        depth = max(level_dict.keys())
        for current_depth in range(depth+1):
            for key in level_dict[current_depth]:
                nsname = f"rail.{key}"
                if current_depth == 0:
                    nsname = f"rail.{key}"
                    cls.TREE[key] = cls.NAMESPACE_MODULE_DICT[key]
                else:
                    parent_key = '.'.join(key.split('.')[0:current_depth])
                    if parent_key in cls.TREE:
                        cls.TREE[parent_key].append({key:cls.NAMESPACE_MODULE_DICT[key]})

        return cls.TREE
        
    @classmethod
    def pretty_print_tree(cls, the_dict=None, indent=""):
        """Utility function to help print the namespace tree

        This can be called recurisvely to walk the tree structure, which has nested dicts

        Parameters
        ----------
        the_dict:  dict | None
            Current dictionary to print, if None it will print cls.TREE
    
        indent:  str
            Indentation string prepended to each line
        """
        if the_dict is None:  # pragma: no cover
            the_dict = cls.TREE
        for key, val in the_dict.items():
            nsname = f"rail.{key}"
            if nsname in cls.PACKAGES:
                pkg_type = "Package"
            else:
                pkg_type = "Namespace"

            print(f"{indent}{pkg_type} {nsname}")
            for vv in val:
                if isinstance(vv, dict):
                    cls.pretty_print_tree(vv, indent=indent+"    ")
                else:
                    print(f"    {indent}{vv.name}")
        return

                    
    @classmethod
    def print_rail_namespace_tree(cls):
        """Print the namespace tree in a nice way"""
        if not cls.TREE:
            cls.build_rail_namespace_tree()
        cls.pretty_print_tree(cls.TREE)
        return
        
    @classmethod
    def do_pkg_api_rst(cls, basedir, key, val):
        """Build the api rst file for a rail package"""
        
        api_pkg_toc = f"rail.{key} package\n"
        api_pkg_toc += "="*len(api_pkg_toc)

        api_pkg_toc += \
f"""
.. automodule:: rail.{key}
    :members:
    :undoc-members:
    :show-inheritance:

Submodules
----------

.. toctree::
    :maxdepth: 4

"""
    
        for vv in val:
            if isinstance(vv, dict):  # pragma: no cover
                for k3, v3 in vv.items():
                    for v4 in v3:
                        api_pkg_toc += f"    {v4.name}.rst\n"
            else:
                api_pkg_toc += f"    {vv.name}.rst\n"                        

        with open(os.path.join(basedir, 'api', f"rail.{key}.rst"), 'w') as apitocfile:
            apitocfile.write(api_pkg_toc)
        return

            
    @classmethod
    def do_namespace_api_rst(cls, basedir, key, val):
        """Build the api rst file for a rail namespace"""

        api_pkg_toc = f"{key} namespace\n"
        api_pkg_toc += "="*len(api_pkg_toc)

        api_pkg_toc += \
"""

.. py:module:: rail.estimation

Subpackages
-----------

.. toctree::
    :maxdepth: 4

{sub_packages}

Submodules
----------

.. toctree::
    :maxdepth: 4

{sub_modules}
"""

        sub_packages = ""
        sub_modules = ""
        for vv in val:
            if isinstance(vv, dict):
                for k3, v3 in vv.items():
                    cls.do_namespace_api_rst(basedir, k3, v3)
                    sub_packages += f"    rail.{k3}\n"
            else:
                sub_modules += f"    {vv.name}\n"
        api_pkg_toc = api_pkg_toc.format(sub_packages=sub_packages, sub_modules=sub_modules)
            
        with open(os.path.join(basedir, 'api', f"rail.{key}.rst"), 'w') as apitocfile:
            apitocfile.write(api_pkg_toc)
        return

            
    @classmethod
    def do_api_rst(cls, basedir='.'):
        if not cls.TREE:  # pragma: no cover
            cls.build_rail_namespace_tree()

        apitoc = \
"""API Documentation
=================

Information on specific functions, classes, and methods.

.. toctree::

"""
        try:
            os.makedirs(basedir)
        except:
            pass

        try:
            os.makedirs(os.path.join(basedir, 'api'))
        except:  # pragma: no cover
            pass

        for key, val in cls.TREE.items():        
            nsname = f"rail.{key}"
            nsfile = os.path.join('api', f"{nsname}.rst")
            apitoc += f"    {nsfile}\n"

            if nsname in cls.PACKAGES:
                cls.do_pkg_api_rst(basedir, key, val)
            else:
                cls.do_namespace_api_rst(basedir, key, val) 

        with open(os.path.join(basedir, 'api.rst'), 'w') as apitocfile:
            apitocfile.write(apitoc)

        return


    @classmethod
    def import_all_packages(cls):
        """Import all the packages that are available in the RAIL ecosystem"""
        pkgs = cls.list_rail_packages()
        for pkg in pkgs.keys():
            try:
                imported_module = importlib.import_module(pkg)
                print(f"Imported {pkg}")
            except Exception as msg:
                print(f"Failed to import {pkg} because: {str(msg)}")


    @classmethod
    def attach_stages(cls, to_module):
        """Attach all the available stages to this module
        
        This allow you to do 'from rail.stages import *'
        """
        from rail.core.stage import RailStage
        cls.STAGE_DICT.clear()
        cls.STAGE_DICT['none'] = []
        cls.BASE_STAGES.clear()
        
        n_base_classes = 0
        n_stages = 0

        for stage_name, stage_info in RailStage.incomplete_pipeline_stages.items():
            if stage_info[0] in [RailStage]:
                continue
            cls.BASE_STAGES.append(stage_info[0])
            cls.STAGE_DICT[stage_info[0].__name__] = []
            n_base_classes += 1

        for stage_name, stage_info in RailStage.pipeline_stages.items():
            setattr(to_module, stage_name, stage_info[0])
            n_stages += 1

        for stage_name, stage_info in RailStage.pipeline_stages.items():
            baseclass = "none"
            for possible_base in cls.BASE_STAGES:
                if issubclass(stage_info[0], possible_base):
                    baseclass = possible_base.__name__
                    break
            cls.STAGE_DICT[baseclass].append(stage_name)

        print(f"Attached {n_base_classes} base classes and {n_stages} fully formed stages to rail.stages")
        return


    @classmethod
    def print_rail_stage_dict(cls):
        """Print an dict of all the RailSages organized by their base class"""
        for key, val in cls.STAGE_DICT.items():
            print(f"BaseClass {key}")
            for vv in val:
                print(f"  {vv}")



"""
Example code that just spits out random numbers between 0 and 3
for z_mode, and Gaussian centered at z_mode with width
random_width*(1+zmode).
"""

import numpy as np
from scipy.stats import norm
from ceci.config import StageParameter as Param
from rail.estimation.estimator import CatEstimator
from rail.core.data import TableHandle
import qp


class RandomGaussEstimator(CatEstimator):
    """Random CatEstimator
    """

    name = 'RandomGaussEstimator'
    inputs = [('input', TableHandle)]
    config_options = CatEstimator.config_options.copy()
    config_options.update(rand_width=Param(float, 0.025, "ad hock width of PDF"),
                          rand_zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
                          rand_zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
                          nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
                          seed=Param(int, 87, msg="random seed"),
                          column_name=Param(str, "mag_i_lsst", msg="name of a column that has the correct number of galaxies to find length of"))

    def __init__(self, args, comm=None):
        """ Constructor:
        Do CatEstimator specific initialization """
        CatEstimator.__init__(self, args, comm=comm)
        self.zgrid = None

    def _process_chunk(self, start, end, data, first):
        pdf = []
        # allow for either format for now
        numzs = len(data[self.config.column_name])
        rng = np.random.default_rng(seed=self.config.seed + start)
        zmode = np.round(rng.uniform(0.0, self.config.rand_zmax, numzs), 3)
        widths = self.config.rand_width * (1.0 + zmode)
        self.zgrid = np.linspace(self.config.rand_zmin, self.config.rand_zmax, self.config.nzbins)
        for i in range(numzs):
            pdf.append(norm.pdf(self.zgrid, zmode[i], widths[i]))
        qp_d = qp.Ensemble(qp.stats.norm, data=dict(loc=np.expand_dims(zmode, -1),  # pylint: disable=no-member
                                                    scale=np.expand_dims(widths, -1)))
        qp_d.set_ancil(dict(zmode=zmode))
        self._do_chunk_output(qp_d, start, end, first)



"""Degrader that applies a cut to given columns."""

from numbers import Number

import numpy as np
from rail.creation.degrader import Degrader


class QuantityCut(Degrader):
    """Degrader that applies a cut to the given columns.

    Note if a galaxy fails any of the cuts on any one of its columns, that
    galaxy is removed from the sample.
    """

    name = "QuantityCut"
    config_options = Degrader.config_options.copy()
    config_options.update(cuts=dict)

    def __init__(self, args, comm=None):
        """
        Constructor

        Does standard Degrader initialization and also gets defines the cuts to be applied
        """
        Degrader.__init__(self, args, comm=comm)
        self.cuts = None
        self.set_cuts(self.config["cuts"])

    def set_cuts(self, cuts: dict):
        """
        Parameters
        ----------
        cuts : dict
            A dictionary of cuts to make on the data.

        Notes
        -----
        The cut keys should be the names of columns you wish to make cuts on.
        The cut values should be either:
        - a number, which is the maximum value. I.e. if the dictionary
        contains "i": 25, then values of i > 25 are cut from the sample.
        - an iterable, which is the range of acceptable values.
        I.e. if the dictionary contains "redshift": (1.5, 2.3), then
        redshifts outside that range are cut from the sample.
        """

        # check that cuts is a dictionary
        if not isinstance(cuts, dict):  # pragma: no cover
            raise TypeError("cuts must be a dictionary.")

        # validate all the cuts and standardize format in dictionary
        self.cuts = {}
        for quantity, cut in cuts.items():
            bad_cut_msg = (
                f"Cut for {quantity} must be a number or an iterable of (min, max)"
            )
            # if single number is provided, save that as the maximum
            if isinstance(cut, Number):
                self.cuts[quantity] = (-np.inf, cut)
            # else, if it's an iterable...
            elif hasattr(cut, "__iter__"):
                # if the iterable is a string or dict, raise error
                if isinstance(cut, (dict, str)):
                    raise TypeError(bad_cut_msg)
                # if the length of the iterable isn't 2, raise error
                if len(cut) != 2:
                    raise ValueError(bad_cut_msg)
                # check that both of the cut values are a number
                for c in cut:
                    if not isinstance(c, Number):
                        raise TypeError(bad_cut_msg)
                # check that the cuts are (min, max)
                if cut[0] > cut[1]:
                    raise ValueError(
                        f"The max for {quantity} must be greater than the min."
                    )
                # if all of these checks passed, save the cuts
                self.cuts[quantity] = (cut[0], cut[1])
            # if the cut isn't a number or iterable, it's the wrong type
            else:
                raise TypeError(bad_cut_msg)

    def run(self):
        """Run method

        Applies cuts

        Notes
        -----
        Get the input data from the data store under this stages 'input' tag
        Puts the data into the data store under this stages 'output' tag
        """
        data = self.get_data("input")

        # get overlap of columns from data and columns on which to make cuts
        columns = set(self.cuts.keys()).intersection(data.columns)

        if len(columns) == 0:  # pragma: no cover
            self.add_data("output", data)
        else:
            # generate a pandas query from the cuts
            query = [
                f"{col} > {self.cuts[col][0]} & {col} < {self.cuts[col][1]}"
                for col in columns
            ]
            query = " & ".join(query)

            out_data = data.query(query)
            self.add_data("output", out_data)

    def __repr__(self):  # pragma: no cover
        """Pretty print this object"""
        printMsg = "Degrader that applies the following cuts to a pandas DataFrame:\n"
        printMsg += "{column: (min, max), ...}\n"
        printMsg += self.cuts.__str__()
        return printMsg



import os
from typing import Type

import numpy as np
import pandas as pd
import pytest

from rail.core.data import DATA_STORE, TableHandle
from rail.core.utilStages import ColumnMapper
from rail.creation.degradation.lsst_error_model import LSSTErrorModel
from rail.creation.degradation.quantityCut import QuantityCut
from rail.creation.degradation.spectroscopic_selections import *


@pytest.fixture
def data():
    """Some dummy data to use below."""

    DS = DATA_STORE()
    DS.__class__.allow_overwrite = True

    # generate random normal data
    rng = np.random.default_rng(0)
    x = rng.normal(loc=26, scale=1, size=(100, 7))

    # replace redshifts with reasonable values
    x[:, 0] = np.linspace(0, 2, x.shape[0])

    # return data in handle wrapping a pandas DataFrame
    df = pd.DataFrame(x, columns=["redshift", "u", "g", "r", "i", "z", "y"])
    return DS.add_data("data", df, TableHandle, path="dummy.pd")


@pytest.fixture
def data_forspec():
    """Some dummy data to use below."""

    DS = DATA_STORE()
    DS.__class__.allow_overwrite = True

    # generate random normal data
    rng = np.random.default_rng(0)
    x = rng.normal(loc=26, scale=1, size=(200000, 7))

    # replace redshifts with reasonable values
    x[:, 0] = np.linspace(0, 2, x.shape[0])

    # return data in handle wrapping a pandas DataFrame
    df = pd.DataFrame(x, columns=["redshift", "u", "g", "r", "i", "z", "y"])
    return DS.add_data("data_forspec", df, TableHandle, path="dummy_forspec.pd")


@pytest.mark.parametrize(
    "cuts,error",
    [
        (1, TypeError),
        ({"u": "cut"}, TypeError),
        ({"u": dict()}, TypeError),
        ({"u": [1, 2, 3]}, ValueError),
        ({"u": [1, "max"]}, TypeError),
        ({"u": [2, 1]}, ValueError),
        ({"u": TypeError}, TypeError),
    ],
)
def test_QuantityCut_bad_params(cuts, error):
    """Test bad parameters that should return Type and Value errors"""
    with pytest.raises(error):
        QuantityCut.make_stage(cuts=cuts)


def test_QuantityCut_returns_correct_shape(data):
    """Make sure QuantityCut is returning the correct shape"""

    cuts = {
        "u": 0,
        "y": (1, 2),
    }
    degrader = QuantityCut.make_stage(cuts=cuts)

    degraded_data = degrader(data).data

    assert degraded_data.shape == data.data.query("u < 0 & y > 1 & y < 2").shape
    os.remove(degrader.get_output(degrader.get_aliased_tag("output"), final_name=True))


@pytest.mark.parametrize(
    "settings,error",
    [
        ({"m5": 1}, TypeError),
        ({"tvis": "xx"}, TypeError),
        ({"nYrObs": "xx"}, TypeError),
        ({"airmass": "xx"}, TypeError),
        ({"extendedSource": "xx"}, TypeError),
        ({"sigmaSys": "xx"}, TypeError),
        ({"magLim": "xx"}, TypeError),
        ({"ndFlag": "xx"}, TypeError),
        ({"tvis": -1}, ValueError),
        ({"nYrObs": -1}, ValueError),
        ({"airmass": -1}, ValueError),
        ({"extendedSource": -1}, ValueError),
        ({"sigmaSys": -1}, ValueError),
        ({"bandNames": False}, TypeError),
        ({"nVisYr": False}, TypeError),
        ({"gamma": False}, TypeError),
        ({"Cm": False}, TypeError),
        ({"msky": False}, TypeError),
        ({"theta": False}, TypeError),
        ({"km": False}, TypeError),
        ({"nVisYr": {}}, ValueError),
        ({"gamma": {}}, ValueError),
        ({"Cm": {}}, ValueError),
        ({"msky": {}}, ValueError),
        ({"theta": {}}, ValueError),
        ({"km": {}}, ValueError),
        ({"nVisYr": {f"{b}": False for b in "ugrizy"}}, TypeError),
        ({"gamma": {f"{b}": False for b in "ugrizy"}}, TypeError),
        ({"Cm": {f"{b}": False for b in "ugrizy"}}, TypeError),
        ({"msky": {f"{b}": False for b in "ugrizy"}}, TypeError),
        ({"theta": {f"{b}": False for b in "ugrizy"}}, TypeError),
        ({"km": {f"{b}": False for b in "ugrizy"}}, TypeError),
        ({"nVisYr": {f"{b}": -1 for b in "ugrizy"}}, ValueError),
        ({"gamma": {f"{b}": -1 for b in "ugrizy"}}, ValueError),
        ({"theta": {f"{b}": -1 for b in "ugrizy"}}, ValueError),
        ({"km": {f"{b}": -1 for b in "ugrizy"}}, ValueError),
    ],
)
def test_LSSTErrorModel_bad_params(settings, error):
    """Test bad parameters that should raise Value and Type errors."""
    with pytest.raises(error):
        LSSTErrorModel.make_stage(**settings)


@pytest.mark.parametrize("m5,highSNR", [({}, False), ({"u": 23}, True)])
def test_LSSTErrorModel_returns_correct_shape(m5, highSNR, data):
    """Test that the LSSTErrorModel returns the correct shape"""

    degrader = LSSTErrorModel.make_stage(m5=m5, highSNR=highSNR)

    degraded_data = degrader(data).data

    assert degraded_data.shape == (data.data.shape[0], 2 * data.data.shape[1] - 1)
    os.remove(degrader.get_output(degrader.get_aliased_tag("output"), final_name=True))


@pytest.mark.parametrize("m5,highSNR", [({}, False), ({"u": 23}, True)])
def test_LSSTErrorModel_magLim(m5, highSNR, data):
    """Test that no mags are above magLim"""

    magLim = 27
    ndFlag = np.nan
    degrader = LSSTErrorModel.make_stage(magLim=magLim, ndFlag=ndFlag, m5=m5, highSNR=highSNR)

    degraded_data = degrader(data).data
    degraded_mags = degraded_data.iloc[:, 1:].to_numpy()

    assert degraded_mags[~np.isnan(degraded_mags)].max() < magLim
    os.remove(degrader.get_output(degrader.get_aliased_tag("output"), final_name=True))


@pytest.mark.parametrize("highSNR", [False, True])
def test_LSSTErrorModel_get_limiting_mags(highSNR):
    degrader = LSSTErrorModel.make_stage(highSNR=highSNR)

    # make sure that the 5-sigma single-visit limiting mags match
    assert np.allclose(
        list(degrader.get_limiting_mags().values()),
        list(degrader._all_m5.values()),
        rtol=1e-4,
    )

    # make sure the coadded mags are deeper
    assert np.all(
        np.array(list(degrader.get_limiting_mags(coadded=True).values()))
        > np.array(list(degrader.get_limiting_mags(coadded=False).values()))
    )

    # test that _get_NSR is the inverse of get_limiting_mags(coadded=True)
    for SNR in [1, 5, 10, 100]:
        limiting_mags = degrader.get_limiting_mags(Nsigma=SNR, coadded=True)
        limiting_mags = np.array(list(limiting_mags.values()))
        NSR = degrader._get_NSR(limiting_mags, degrader.config["bandNames"].keys())
        assert np.allclose(1 / NSR, SNR)


def test_LSSTErrorModel_extended(data):
    n_samples = len(data.data)
    rng = np.random.default_rng(0)
    data.data["major"] = np.abs(rng.normal(loc=0.01, scale=0.1, size=n_samples))  # add major and minor axes
    b_to_a = 1 - 0.5 * rng.random(n_samples)
    data.data["minor"] = data.data["major"] * b_to_a

    errorModel_auto = LSSTErrorModel.make_stage(name="error_model_auto", errorType="auto")
    errorModel_auto(data)
    errorModel_auto.__repr__()

    errorModel_gaap = LSSTErrorModel.make_stage(name="error_model_gaap", errorType="gaap")
    errorModel_gaap(data)
    errorModel_gaap.__repr__()

    os.remove(errorModel_auto.get_output(errorModel_auto.get_aliased_tag("output"), final_name=True))
    os.remove(errorModel_gaap.get_output(errorModel_gaap.get_aliased_tag("output"), final_name=True))


@pytest.mark.parametrize("errType, errortype", [("sakana", ValueError)])
def test_LSSTErrorModel_bad_type(errType, errortype):
    """Test bad parameters that should raise ValueError"""
    with pytest.raises(errortype):
        LSSTErrorModel.make_stage(errorType=errType)


@pytest.mark.parametrize(
    "degrader",
    [
        LSSTErrorModel.make_stage(),
        LSSTErrorModel.make_stage(highSNR=True),
    ],
)
def test_random_seed(degrader, data):
    """Test control with random seeds."""

    # make sure setting the same seeds yields the same output
    degraded_data1 = degrader(data, seed=0).data
    degraded_data2 = degrader(data, seed=0).data
    assert degraded_data1.equals(degraded_data2)

    # make sure setting different seeds yields different output
    degraded_data3 = degrader(data, seed=1).data.to_numpy()
    assert not degraded_data1.equals(degraded_data3)
    os.remove(degrader.get_output(degrader.get_aliased_tag("output"), final_name=True))


def test_SpecSelection(data):
    bands = ["u", "g", "r", "i", "z", "y"]
    band_dict = {band: f"mag_{band}_lsst" for band in bands}
    rename_dict = {f"{band}_err": f"mag_err_{band}_lsst" for band in bands}
    rename_dict.update({f"{band}": f"mag_{band}_lsst" for band in bands})
    standard_colnames = [f"mag_{band}_lsst" for band in "ugrizy"]

    col_remapper_test = ColumnMapper.make_stage(
        name="col_remapper_test", hdf5_groupname="", columns=rename_dict
    )
    data = col_remapper_test(data)

    degrader_GAMA = SpecSelection_GAMA.make_stage()
    degrader_GAMA(data)
    degrader_GAMA.__repr__()

    os.remove(degrader_GAMA.get_output(degrader_GAMA.get_aliased_tag("output"), final_name=True))

    degrader_BOSS = SpecSelection_BOSS.make_stage()
    degrader_BOSS(data)
    degrader_BOSS.__repr__()

    os.remove(degrader_BOSS.get_output(degrader_BOSS.get_aliased_tag("output"), final_name=True))

    degrader_DEEP2 = SpecSelection_DEEP2.make_stage()
    degrader_DEEP2(data)
    degrader_DEEP2.__repr__()

    os.remove(degrader_DEEP2.get_output(degrader_DEEP2.get_aliased_tag("output"), final_name=True))

    degrader_VVDSf02 = SpecSelection_VVDSf02.make_stage()
    degrader_VVDSf02(data)
    degrader_VVDSf02.__repr__()

    degrader_zCOSMOS = SpecSelection_zCOSMOS.make_stage(colnames={"i": "mag_i_lsst", "redshift": "redshift"})
    degrader_zCOSMOS(data)
    degrader_zCOSMOS.__repr__()

    os.remove(degrader_zCOSMOS.get_output(degrader_zCOSMOS.get_aliased_tag("output"), final_name=True))

    degrader_HSC = SpecSelection_HSC.make_stage()
    degrader_HSC(data)
    degrader_HSC.__repr__()

    os.remove(degrader_HSC.get_output(degrader_HSC.get_aliased_tag("output"), final_name=True))

    degrader_HSC = SpecSelection_HSC.make_stage(percentile_cut=70)
    degrader_HSC(data)
    degrader_HSC.__repr__()

    os.remove(degrader_HSC.get_output(degrader_HSC.get_aliased_tag("output"), final_name=True))


def test_SpecSelection_low_N_tot(data_forspec):
    bands = ["u", "g", "r", "i", "z", "y"]
    band_dict = {band: f"mag_{band}_lsst" for band in bands}
    rename_dict = {f"{band}_err": f"mag_err_{band}_lsst" for band in bands}
    rename_dict.update({f"{band}": f"mag_{band}_lsst" for band in bands})
    standard_colnames = [f"mag_{band}_lsst" for band in "ugrizy"]

    col_remapper_test = ColumnMapper.make_stage(
        name="col_remapper_test", hdf5_groupname="", columns=rename_dict
    )
    data_forspec = col_remapper_test(data_forspec)

    degrader_zCOSMOS = SpecSelection_zCOSMOS.make_stage(N_tot=1)
    degrader_zCOSMOS(data_forspec)

    os.remove(degrader_zCOSMOS.get_output(degrader_zCOSMOS.get_aliased_tag("output"), final_name=True))


@pytest.mark.parametrize("N_tot, errortype", [(-1, ValueError)])
def test_SpecSelection_bad_params(N_tot, errortype):
    """Test bad parameters that should raise TypeError"""
    with pytest.raises(errortype):
        SpecSelection.make_stage(N_tot=N_tot)


@pytest.mark.parametrize("errortype", [(ValueError)])
def test_SpecSelection_bad_colname(data, errortype):
    """Test bad parameters that should raise TypeError"""
    with pytest.raises(errortype):
        degrader_GAMA = SpecSelection_GAMA.make_stage()
        degrader_GAMA(data)


@pytest.mark.parametrize("success_rate_dir, errortype", [("/this/path/should/not/exist", ValueError)])
def test_SpecSelection_bad_path(success_rate_dir, errortype):
    """Test bad parameters that should raise TypeError"""
    with pytest.raises(errortype):
        SpecSelection.make_stage(success_rate_dir=success_rate_dir)





class MetricEvaluator:
    """ A superclass for metrics evaluations"""
    def __init__(self, qp_ens):
        """Class constructor.
        Parameters
        ----------
        qp_ens: qp.Ensemble object
            PDFs as qp.Ensemble
        """
        self._qp_ens = qp_ens

    def evaluate(self):  #pragma: no cover
        """
        Evaluates the metric a function of the truth and prediction

        Returns
        -------
        metric: dictionary
            value of the metric and statistics thereof
        """
        raise NotImplementedError



"""The LSST Model for photometric errors."""

from numbers import Number
from typing import Iterable, List, Optional, Tuple

import numpy as np
from rail.creation.degrader import Degrader


class LSSTErrorModel(Degrader):
    """The LSST Model for photometric errors.

    Implements the error model from the LSST Overview Paper:
    https://arxiv.org/abs/0805.2366
    Note however that this paper gives the high SNR approximation.
    By default, this model uses the more accurate version of the error model
    where Eq. 5 = (N/S)^2, in flux, and the error is Gaussian in flux space.
    There is a flag allowing you to use the high SNR approximation instead.
    See the __init__ docstring.

    In addition, this paper only gives the error estimation for point sources.
    To account for extended elliptical sources, we follow:
    http://arxiv.org/abs/2007.01846
    We apply two methods to estimate the aperture sizes of the sources according
    to the definitation of measured fluxes:
    1) 'auto' model given by http://arxiv.org/abs/2007.01846;
    2) 'gaap' model given by https://arxiv.org/abs/1902.11265
    In order to use these methods, your dataframe of galaxy data must include columns
    named "major" and "minor", which are the semi-major and minor axes of the galaxies
    in arcseconds. For the cosmoDC2 catalog, they are the half-light radii.

    Create an instance by calling the class, then use the instance as a
    callable on pandas DataFrames.
    Example usage:
    errModel = LSSTErrorModel()
    data_with_errs = errModel(data)
    Error model from the LSST Overview Paper:
    https://arxiv.org/abs/0805.2366

    All parameters are optional. To see the default settings, do
    `LSSTErrorModel().config_options()`

    Note that the dictionary bandNames sets the bands for which this model
    calculates photometric errors. The dictionary keys are the band names
    that the error model uses internally to search for parameters, and the
    corresponding dictionary values are the band names as they appear in
    your data set. By default, the LSST bands are named "u", "g", "r", "i", "z",
    and "y". You can use the bandNames dictionary to alias them differently.

    For example, if in your DataFrame, the bands are named lsst_u, lsst_g, etc.
    you can set bandNames = {"u": "lsst_u", "g": "lsst_g", ...},
    and the error model will work automatically.
    You can also add other bands to bandNames. For example, if you want to
    use the same model to calculate photometric errors for Euclid bands, you
    can include {"euclid_y": "euclid_y", "euclid_j": "euclid_j", ...}.
    In this case, you must include the additional information listed below...

    IMPORTANT: For every band in bandNames, you must provide:
        - nVisYr
        - gamma
        - the single-visit 5-sigma limiting magnitude. You can do this either by
          (1) explicitly providing it in the m5 dictionary, or
          (2) by adding the corresponding parameters to Cm, msky, theta,
          and km, in which case the limiting magnitude will be calculated
          for you, using Eq. 6 from the LSST Overview Paper.

    Note if for any bands, you explicitly pass a limiting magnitude in the
    m5 dictionary, the model will use the explicitly passed value,
    regardless of the values in Cm, msky, theta, and km.

    Parameters
    ----------
    bandNames : dict, optional
        A dictionary of bands for which to calculate errors. The dictionary
        keys are the band names that the Error Model uses internally to
        search for parameters, and the corresponding dictionary values
        are the names of those bands as they appear in your data set.
        Can be used to alias the default names of the LSST bands, or to add
        additional bands. See notes above.
    tvis : float, optional
        Exposure time for a single visit
    nYrObs : float, optional
        Number of years of observations
    nVisYr : dict, optional
        Mean number of visits per year in each band
    gamma : dict, optional
        A band dependent parameter defined in the LSST Overview Paper
    airmass : float, optional
        The fiducial airmass
    extendedSource : float, optional
        Constant to add to magnitudes of extended sources.
        The error model is designed to emulate magnitude errors for point
        sources. This constant provides a zeroth order correction accounting
        for the fact that extended sources have larger uncertainties. Note
        this is only meant to account for small, slightly extended sources.
        For typical LSST galaxies, this may be of order ~0.3.
    sigmaSys : float, optional
        The irreducible error of the system in AB magnitudes.
        Set's the minimum photometric error.
    magLim : float, optional
        The dimmest magnitude allowed. All dimmer magnitudes are set to ndFlag.
    ndFlag : float, optional
        The flag for non-detections. All magnitudes greater than magLim (and
        their corresponding errors) will be set to this value.
    m5 : dict, optional
        A dictionary of single visit 5-sigma limiting magnitudes. For any
        bands for which you pass a value in m5, this will be the 5-sigma
        limiting magnitude used, and any values for that band in Cm, msky,
        theta, and km will be ignored.
    Cm : dict, optional
        A band dependent parameter defined in the LSST Overview Paper
    msky : dict, optional
        Median zenith sky brightness in each band
    theta : dict, optional
        Median zenith seeing FWHM (in arcseconds) for each band
    km : dict, optional
        Atmospheric extinction in each band
    highSNR : bool, default=False
        Sets whether you use the high SNR approximation given in the LSST
        Overview Paper. If False, then Eq. 5 from the LSST Error Model is
        used to calculate (N/S)^2 in flux, and errors are Gaussian in flux
        space. If True, then Eq. 5 is used to calculate the squared error
        in magnitude space, and errors are Gaussian in magnitude space.
    A_min : float, default=2.0
        The minimum GAaP aperture size (in arcmin).
    A_max : float, default=0.7
        The maximum GAaP aperture size (in arcmin).
    errorType: string, default="point"
        Should be "point" for point source errors. For errors for extended sources,
        you can choose either "auto" or "gaap". See the top of the docstring for
        more details.
    """

    name = "LSSTErrorModel"
    config_options = Degrader.config_options.copy()
    config_options.update(
        **{
            "bandNames": {  # provided so you can alias the names of the bands
                "u": "u",
                "g": "g",
                "r": "r",
                "i": "i",
                "z": "z",
                "y": "y",
            },
            "tvis": 30.0,  # exposure time for a single visit in seconds
            "nYrObs": 10.0,  # number of years of observations
            "nVisYr": {  # mean number of visits per year in each filter
                "u": 5.6,
                "g": 8.0,
                "r": 18.4,
                "i": 18.4,
                "z": 16.0,
                "y": 16.0,
            },
            "gamma": {  # band dependent parameter
                "u": 0.038,
                "g": 0.039,
                "r": 0.039,
                "i": 0.039,
                "z": 0.039,
                "y": 0.039,
            },
            "airmass": 1.2,  # fiducial airmass
            "extendedSource": 0.0,  # constant added to m5 for extended sources
            "sigmaSys": 0.005,  # expected irreducible error
            "magLim": 30.0,  # dimmest allowed magnitude; dimmer mags set to ndFlag
            "ndFlag": np.nan,  # flag for non-detections (all mags > magLim)
            "m5": {},  # explicit list of m5 limiting magnitudes
            "Cm": {  # band dependent parameter
                "u": 23.09,
                "g": 24.42,
                "r": 24.44,
                "i": 24.32,
                "z": 24.16,
                "y": 23.73,
            },
            "msky": {  # median zenith sky brightness at Cerro Pachon
                "u": 22.99,
                "g": 22.26,
                "r": 21.20,
                "i": 20.48,
                "z": 19.60,
                "y": 18.61,
            },
            "theta": {  # median zenith seeing FWHM, arcseconds
                "u": 0.81,
                "g": 0.77,
                "r": 0.73,
                "i": 0.71,
                "z": 0.69,
                "y": 0.68,
            },
            "km": {  # atmospheric extinction
                "u": 0.491,
                "g": 0.213,
                "r": 0.126,
                "i": 0.096,
                "z": 0.069,
                "y": 0.170,
            },
            "highSNR": False,
            "A_min": 0.7,  # minimum aperture size in arcseconds
            "A_max": 2.0,  # maximum aperture size in arcseconds
            "errorType": "point",
        }
    )

    def __init__(self, args, comm=None):
        Degrader.__init__(self, args, comm=comm)

        # validate the settings
        self._validate_settings()

        # calculate the single-visit 5-sigma limiting magnitudes using the settings
        self._all_m5 = self._calculate_m5()

        # update the limiting magnitudes with any m5s passed
        if self.config.m5 is not None:
            self._all_m5.update(self.config.m5)

    def _validate_settings(self):
        """Validate all the settings."""

        # check that highSNR is boolean
        if not isinstance(self.config["highSNR"], bool):  # pragma: no cover
            raise TypeError("highSNR must be boolean.")

        if self.config["errorType"] not in ["point", "gaap", "auto"]:
            raise ValueError("errorType should be one of 'point', 'gaap', 'auto'")

        # check all the numbers
        for key in [
            "tvis",
            "nYrObs",
            "airmass",
            "extendedSource",
            "sigmaSys",
            "magLim",
            "ndFlag",
            "A_min",
            "A_max",
        ]:
            # check they are numbers
            # note we also check if they're bools and np.nan's because these
            # are both technically numbers
            is_number = isinstance(self.config[key], Number)
            is_bool = isinstance(self.config[key], bool)
            is_nan = np.isnan(self.config[key])
            # ndFlag can be np.nan
            if key == "ndFlag":
                if not (is_number or is_nan) or is_bool:  # pragma: no cover
                    raise TypeError(f"{key} must be a number or NaN.")
            # the others cannot
            else:
                if not is_number or is_nan or is_bool:  # pragma: no cover
                    raise TypeError(f"{key} must be a number.")
            # if they are numbers, check that they are non-negative
            # except for magLim and ndFlag, which can be
            if key not in ["magLim", "ndFlag"]:
                if self.config[key] < 0:
                    raise ValueError(f"{key} must be non-negative.")

        # make sure bandNames is a dictionary
        if not isinstance(self.config["bandNames"], dict):  # pragma: no cover
            raise TypeError(
                "bandNames must be a dictionary where the keys are the names "
                "of the bands as used internally by the error model, and the "
                "values are the names of the bands as present in your data set."
            )

        # remove unnecessary keys from all setting dictionaries
        for key, val in self.config.items():
            if key in ["aliases"]:
                continue
            if isinstance(val, dict):
                self.config[key] = {
                    band: val[band] for band in self.config["bandNames"] if band in val
                }

        # check all the other dictionaries
        for key in ["nVisYr", "gamma", "Cm", "msky", "theta", "km"]:

            # make sure they are dictionaries
            if not isinstance(self.config[key], dict):  # pragma: no cover
                raise TypeError(f"{key} must be a dictionary.")

            # check the values in the dictionary
            for subkey, val in self.config[key].items():

                # check that it's a number
                is_number = isinstance(val, Number)
                is_bool = isinstance(val, bool)
                if not is_number or is_bool:
                    raise TypeError(f"{key}['{subkey}'] must be a number.")

                # for certain dictionaries, check that the numbers are positive
                if key not in ["Cm", "msky"]:
                    if val < 0:
                        raise ValueError(f"{key}['{subkey}'] must be positive.")

            # get the set of bands in bandNames that aren't in this dictionary
            missing = set(self.config["bandNames"]) - set(self.config[key])

            # nVisYr and gamma must have an entry for every band in bandNames
            if key in ["nVisYr", "gamma"]:
                if len(missing) > 0:
                    raise ValueError(
                        f"{key} must have an entry for every band in bandNames, "
                        f"and is currently missing entries for {missing}."
                    )

            # but for the other dictionaries...
            else:
                # we don't need an entry for every band in bandNames, as long
                # as the missing bands are listed in m5
                missing -= set(self.config["m5"])

                if len(missing) > 0:
                    raise ValueError(
                        "You haven't provided enough information to calculate "
                        f"limiting magnitudes for {missing}. Please include "
                        f"entries for these bands in {key}, or explicitly pass "
                        "their 5-sigma limiting magnitudes in m5."
                    )

    def _calculate_m5(self) -> dict:
        """Calculate the single-visit m5 limiting magnitudes.

        Uses Eq. 6 from https://arxiv.org/abs/0805.2366
        Note this is only done for the bands for which an m5 wasn't explicitly passed.
        """

        # get the settings
        settings = self.config

        # get the list of bands for which an m5 wasn't explicitly passed
        bands = set(self.config["bandNames"]) - set(self.config["m5"])
        bands = [band for band in self.config["bandNames"] if band in bands]

        # calculate the m5 limiting magnitudes using Eq. 6
        m5 = {
            band: settings["Cm"][band]
            + 0.50 * (self.config["msky"][band] - 21)
            + 2.5 * np.log10(0.7 / settings["theta"][band])
            + 1.25 * np.log10(settings["tvis"] / 30)
            - settings["km"][band] * (settings["airmass"] - 1)
            - settings["extendedSource"]
            for band in bands
        }

        return m5

    def _get_bands_and_names(
        self, columns: Iterable[str]
    ) -> Tuple[List[str], List[str]]:
        """Get the bands and bandNames that are present in the given data columns."""

        # get the list of bands present in the data
        bandNames = list(set(self.config["bandNames"].values()).intersection(columns))

        # sort bandNames to be in the same order provided in settings["bandNames"]
        bandNames = [
            band for band in self.config["bandNames"].values() if band in bandNames
        ]

        # get the internal names of the bands from bandNames
        bands = [
            {bandName: band for band, bandName in self.config["bandNames"].items()}[
                bandName
            ]
            for bandName in bandNames
        ]

        return bands, bandNames

    def _get_area_ratio_gaap(
        self, majors: np.ndarray, minors: np.ndarray, bands: list
    ) -> np.ndarray:
        """Get the ratio between PSF area and galaxy aperture area for "gaap" model."""
        # get the psf size for each band
        theta_size = np.array([self.config["theta"][band] for band in bands])

        hl_to_sigma = 1 / 0.6745  # half IQR to Gaussian sigma
        fwhm_to_sigma = 1 / 2.355

        # convert PSF FWHM to a Gaussian sigma
        theta_sigma = theta_size * fwhm_to_sigma
        A_min_sigma = self.config["A_min"] * fwhm_to_sigma
        A_max_sigma = self.config["A_max"] * fwhm_to_sigma

        # calculate the area of the psf in each band
        A_psf = np.pi * theta_sigma**2

        # convert the half-light radii to the sigma of semi-major and minor axis
        majors *= hl_to_sigma
        minors *= hl_to_sigma

        # calculate the area of the galaxy aperture in each band
        a_ap = np.sqrt(
            theta_sigma[None, :] ** 2 + majors[:, None] ** 2 + A_min_sigma**2
        )
        a_ap[a_ap > A_max_sigma] = A_max_sigma
        b_ap = np.sqrt(
            theta_sigma[None, :] ** 2 + minors[:, None] ** 2 + A_min_sigma**2
        )
        b_ap[b_ap > A_max_sigma] = A_max_sigma
        A_ap = np.pi * a_ap * b_ap

        # return their ratio
        return A_ap / A_psf

    def _get_area_ratio_auto(
        self, majors: np.ndarray, minors: np.ndarray, bands: list
    ) -> np.ndarray:
        """Get the ratio between PSF area and galaxy aperture area for "auto" model."""
        # get the psf size for each band
        theta_size = np.array([self.config["theta"][band] for band in bands])
        fwhm_to_sigma = 1 / 2.355

        # convert PSF FWHM to a Gaussian sigma
        theta_sigma = theta_size * fwhm_to_sigma
        # calculate the area of the psf in each band
        A_psf = np.pi * theta_sigma**2

        # calculate the area of the galaxy aperture in each band
        a_ap = np.sqrt(theta_size[None, :] ** 2 + (2.5 * majors[:, None]) ** 2)
        b_ap = np.sqrt(theta_size[None, :] ** 2 + (2.5 * minors[:, None]) ** 2)
        A_ap = np.pi * a_ap * b_ap

        # return their ratio
        return A_ap / A_psf

    def _get_NSR(self, mags: np.ndarray, bands: list) -> np.ndarray:
        """Calculate the noise-to-signal ratio.

        Uses Eqs 4 and 5 from https://arxiv.org/abs/0805.2366
        """

        # get the 5-sigma limiting magnitudes for these bands
        m5 = np.array([self._all_m5[band] for band in bands])
        # and the values for gamma
        gamma = np.array([self.config["gamma"][band] for band in bands])

        # calculate x as defined in the paper
        x = 10 ** (0.4 * np.subtract(mags, m5))

        # calculate the squared NSR for a single visit
        # Eq. 5 in https://arxiv.org/abs/0805.2366
        nsrRandSqSingleExp = (0.04 - gamma) * x + gamma * x**2

        # calculate the random NSR for the stacked image
        nVisYr = np.array([self.config["nVisYr"][band] for band in bands])
        nStackedObs = nVisYr * self.config["nYrObs"]
        nsrRand = np.sqrt(nsrRandSqSingleExp / nStackedObs)

        # get the irreducible system NSR
        if self.config["highSNR"]:
            nsrSys = self.config["sigmaSys"]
        else:
            nsrSys = 10 ** (self.config["sigmaSys"] / 2.5) - 1

        # calculate the total NSR
        nsr = np.sqrt(nsrRand**2 + nsrSys**2)

        return nsr

    def _get_obs_and_errs(
        self,
        mags: np.ndarray,
        bands: list,
        seed: Optional[int],
        majors: np.ndarray = None,
        minors: np.ndarray = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Return observed magnitudes and magnitude errors."""

        rng = np.random.default_rng(seed)

        # get the aperture-psf ratio for all the galaxies
        if self.config["errorType"] == "gaap":
            a_ratio = self._get_area_ratio_gaap(majors, minors, bands)
        elif self.config["errorType"] == "auto":
            a_ratio = self._get_area_ratio_auto(majors, minors, bands)
        else:
            a_ratio = 1

        # get the NSR for all the galaxies
        nsr = self._get_NSR(mags, bands) * np.sqrt(a_ratio)

        if self.config["highSNR"]:
            # in the high SNR approximation, mag err ~ nsr, and we can
            # model errors as Gaussian in magnitude space

            # calculate observed magnitudes
            obsMags = rng.normal(loc=mags, scale=nsr)

            # decorrelate the magnitude errors from the true magnitudes
            obsMagErrs = self._get_NSR(obsMags, bands) * np.sqrt(a_ratio)

        else:
            # in the more accurate error model, we acknowledge err != nsr,
            # and we model errors as Gaussian in flux space

            # calculate observed magnitudes
            fluxes = 10 ** (mags / -2.5)
            obsFluxes = fluxes * (1 + rng.normal(scale=nsr))
            obsFluxes = np.clip(obsFluxes, 0, None)
            with np.errstate(divide="ignore"):
                obsMags = -2.5 * np.log10(obsFluxes)

            # decorrelate the magnitude errors from the true magnitudes
            obsFluxNSR = self._get_NSR(obsMags, bands) * np.sqrt(a_ratio)
            obsMagErrs = 2.5 * np.log10(1 + obsFluxNSR)

        # flag magnitudes beyond magLim as non-detections
        idx = np.where(obsMags > self.config["magLim"])
        obsMags[idx] = self.config["ndFlag"]
        obsMagErrs[idx] = self.config["ndFlag"]

        return obsMags, obsMagErrs

    def run(self):
        """Calculate errors for data, and save the results in a pandas DataFrame."""
        # get the bands and bandNames present in the data
        data = self.get_data("input", allow_missing=True)
        bands, bandNames = self._get_bands_and_names(data.columns)

        # get numpy array of magnitudes
        mags = data[bandNames].to_numpy()

        if self.config["errorType"] != "point":
            # get the sizes of the galaxies
            majors = data["major"].to_numpy()
            minors = data["minor"].to_numpy()

            # get observed magnitudes and magnitude errors
            obsMags, obsMagErrs = self._get_obs_and_errs(
                mags, bands, self.config.seed, majors, minors
            )
        else:
            # get observed magnitudes and magnitude errors
            obsMags, obsMagErrs = self._get_obs_and_errs(mags, bands, self.config.seed)

        # save the observations in a DataFrame
        obsData = data.copy()
        obsData[bandNames] = obsMags
        obsData[[band + "_err" for band in bandNames]] = obsMagErrs

        # re-order columns so that the error columns come right after the
        # respective magnitudes
        columns = data.columns.tolist()
        for band in bandNames:
            columns.insert(columns.index(band) + 1, band + "_err")
        obsData = obsData[columns]
        self.add_data("output", obsData)

    def get_limiting_mags(
        self,
        Nsigma: float = 5,
        coadded: bool = False,
    ) -> dict:
        """Return the limiting magnitudes for all bands.

        This method essentially reverse engineers the _get_NSR() method
        so that we calculate what magnitude results in NSR = 1/Nsigma.
        (NSR is noise-to-signal ratio; NSR = 1/SNR)

        Parameters
        ----------
        Nsigma : float, default=5
            Sets which limiting magnitude to return, e.g. Nsigma = 1 returns
            the 1-sigma limiting magnitude. In other words, Nsigma is equal
            to the signal-to-noise ratio (SNR) of the limiting magnitudes.
        coadded : bool, default=False
            Whether to return the limiting magnitudes for a single visit,
            or for a coadded image.
        """

        # get the bands and bandNames
        bands, bandNames = self._get_bands_and_names(self.config["bandNames"].values())
        # get the 5-sigma limiting magnitudes for these bands
        m5 = np.array([self._all_m5[band] for band in bands])

        # and the values for gamma
        gamma = np.array([self.config["gamma"][band] for band in bands])

        # get the number of exposures
        if coadded:
            nVisYr = np.array([self.config["nVisYr"][band] for band in bands])
            nStackedObs = nVisYr * self.config["nYrObs"]
        else:
            nStackedObs = 1

        # get the irreducible system error
        if self.config["highSNR"]:
            nsrSys = self.config["sigmaSys"]
        else:
            nsrSys = 10 ** (self.config["sigmaSys"] / 2.5) - 1

        # calculate the square of the random NSR that a single exposure must have
        nsrRandSqSingleExp = (1 / Nsigma**2 - nsrSys**2) * nStackedObs

        # calculate the value of x that corresponds to this NSR
        # note this is just the quadratic equation,
        # applied to NSR^2 = (0.04 - gamma) * x + gamma * x^2
        x = (
            (gamma - 0.04)
            + np.sqrt((gamma - 0.04) ** 2 + 4 * gamma * nsrRandSqSingleExp)
        ) / (2 * gamma)

        # convert x to a limiting magnitude
        limiting_mags = m5 + 2.5 * np.log10(x)

        # return as a dictionary
        return dict(zip(bandNames, limiting_mags))

    def __repr__(self):  # pragma: no cover
        """Define how the model is represented and printed."""

        settings = self.config

        # start message
        printMsg = "LSSTErrorModel parameters:\n\n"

        # list all bands
        printMsg += "Model for bands: "
        printMsg += ", ".join(settings["bandNames"].values()) + "\n"

        # print whether using the high SNR approximation
        if self.config["highSNR"]:
            printMsg += "Using the high SNR approximation\n\n"
        else:
            printMsg += "\n"

        # print which error mode we are using
        printMsg += f"Using error type {settings['errorType']}\n"

        # exposure time
        printMsg += f"Exposure time = {settings['tvis']} s\n"
        # number of years
        printMsg += f"Number of years of observations = {settings['nYrObs']}\n"
        # mean visits per year
        printMsg += "Mean visits per year per band:\n   "
        printMsg += (
            ", ".join(
                [
                    f"{bandName}: {settings['nVisYr'][band]}"
                    for band, bandName in settings["bandNames"].items()
                ]
            )
            + "\n"
        )
        # airmass
        printMsg += f"Airmass = {settings['airmass']}\n"
        # irreducible error
        printMsg += f"Irreducible system error = {settings['sigmaSys']}\n"
        # extended sources
        if settings["extendedSource"] > 0:
            printMsg += f"Extended source model: add {settings['extendedSource']}"
            printMsg += "mag to 5-sigma depth for point sources\n"
        # minimum magnitude
        printMsg += f"Magnitudes dimmer than {settings['magLim']} are "
        printMsg += f"set to {settings['ndFlag']}\n"
        # gamma
        printMsg += "gamma for each band:\n   "
        printMsg += (
            ", ".join(
                [
                    f"{bandName}: {settings['gamma'][band]}"
                    for band, bandName in settings["bandNames"].items()
                ]
            )
            + "\n\n"
        )

        # coadded m5's
        printMsg += "The coadded 5-sigma limiting magnitudes are:\n"
        printMsg += (
            ", ".join(
                [
                    f"{bandName}: {limiting_mag:.2f}"
                    for bandName, limiting_mag in self.get_limiting_mags(
                        coadded=True
                    ).items()
                ]
            )
            + "\n\n"
        )

        # explicit m5
        if len(settings["m5"]) > 0:
            printMsg += "The following single-visit 5-sigma limiting magnitudes "
            printMsg += "were explicitly passed:\n   "
            printMsg += (
                ", ".join(
                    [
                        f"{bandName}: {settings['m5'][band]}"
                        for band, bandName in settings["bandNames"].items()
                        if band in settings["m5"]
                    ]
                )
                + "\n\n"
            )

        m5 = self._calculate_m5()
        if len(m5) > 0:
            # Calculated m5
            printMsg += "The following single-visit 5-sigma limiting magnitudes are\n"
            printMsg += "calculated using the parameters that follow them:\n   "
            printMsg += (
                ", ".join(
                    [
                        f"{settings['bandNames'][band]}: {val:.2f}"
                        for band, val in m5.items()
                    ]
                )
                + "\n"
            )
            # Cm
            printMsg += "Cm for each band:\n   "
            printMsg += (
                ", ".join(
                    [
                        f"{settings['bandNames'][band]}: {val}"
                        for band, val in settings["Cm"].items()
                    ]
                )
                + "\n"
            )
            # msky
            printMsg += "Median zenith sky brightness in each band:\n   "
            printMsg += (
                ", ".join(
                    [
                        f"{settings['bandNames'][band]}: {val}"
                        for band, val in settings["msky"].items()
                    ]
                )
                + "\n"
            )
            # theta
            printMsg += "Median zenith seeing FWHM (in arcseconds) for each band:\n   "
            printMsg += (
                ", ".join(
                    [
                        f"{settings['bandNames'][band]}: {val}"
                        for band, val in settings["theta"].items()
                    ]
                )
                + "\n"
            )
            # km
            printMsg += "Extinction coefficient for each band:\n   "
            printMsg += (
                ", ".join(
                    [
                        f"{settings['bandNames'][band]}: {val}"
                        for band, val in settings["km"].items()
                    ]
                )
                + "\n"
            )

        return printMsg



"""
Abstract base classes defining redshift estimations Informers and Estimators
"""
from rail.core.data import QPHandle, TableHandle, ModelHandle
from rail.core.stage import RailStage


class CatSummarizer(RailStage):  #pragma: no cover
    """The base class for classes that go from catalog-like tables
    to ensemble NZ estimates.

    CatSummarizer take as "input" a catalog-like table.  I.e., a
    table with fluxes in photometric bands among the set of columns.

    provide as "output" a QPEnsemble, with per-ensemble n(z).
    """

    name = 'CatSummarizer'
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=10000)
    inputs = [('input', TableHandle)]
    outputs = [('output', QPHandle)]

    def __init__(self, args, comm=None):
        """Initialize Summarizer"""
        RailStage.__init__(self, args, comm=comm)

    def summarize(self, input_data):
        """The main run method for the summarization, should be implemented
        in the specific subclass.

        This will attach the input_data to this `CatSummarizer`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this `CatSummarizer`
        by using `self.add_data('output', output_data)`.

        Finally, this will return a QPHandle providing access to that output data.

        Parameters
        ----------
        input_data : `dict` or `ModelHandle`
            Either a dictionary of all input data or a `ModelHandle` providing access to the same

        Returns
        -------
        output: `qp.Ensemble`
            Ensemble with n(z), and any ancilary data
        """
        self.set_data('input', input_data)
        self.run()
        self.finalize()
        return self.get_handle('output')


class PZSummarizer(RailStage):
    """The base class for classes that go from per-galaxy PZ estimates to ensemble NZ estimates

    PZSummarizer take as "input" a `qp.Ensemble` with per-galaxy PDFs, and
    provide as "output" a QPEnsemble, with per-ensemble n(z).
    """

    name = 'PZtoNZSummarizer'
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=10000)
    inputs = [('input', QPHandle)]
    outputs = [('output', QPHandle)]

    def __init__(self, args, comm=None):
        """Initialize Estimator that can sample galaxy data."""
        RailStage.__init__(self, args, comm=comm)

    def summarize(self, input_data):
        """The main run method for the summarization, should be implemented
        in the specific subclass.

        This will attach the input_data to this `PZtoNZSummarizer`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Finally, this will return a QPHandle providing access to that output data.

        Parameters
        ----------
        input_data : `qp.Ensemble`
            Per-galaxy p(z), and any ancilary data associated with it

        Returns
        -------
        output: `qp.Ensemble`
            Ensemble with n(z), and any ancilary data
        """
        self.set_data('input', input_data)
        self.run()
        self.finalize()
        return self.get_handle('output')


class SZPZSummarizer(RailStage):  #pragma: no cover
    """The base class for classes that use two sets of data: a photometry sample with
    spec-z values, and a photometry sample with unknown redshifts, e.g. minisom_som and
    outputs a QP Ensemble with bootstrap realization of the N(z) distribution
    """
    name = 'SZPZtoNZSummarizer'
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=10000)
    inputs = [('input', TableHandle),
              ('spec_input', TableHandle),
              ('model', ModelHandle)]
    outputs = [('output', QPHandle)]

    def __init__(self, args, comm=None):
        """Initialize Estimator that can sample galaxy data."""
        RailStage.__init__(self, args, comm=comm)
        self.model = None
        if not isinstance(args, dict):  #pragma: no cover
            args = vars(args)
        self.open_model(**args)

    def open_model(self, **kwargs):
        """Load the mode and/or attach it to this Summarizer

        Notes
        -----
        Keywords are potentially used to build the model.
        By default we just use 'model'  which can be an `object`, `str` or `ModelHandle`

        I.e., either an object with a trained model, a path pointing to a file that
        can be read to obtain the trained model, or a `ModelHandle` providing access to the trained model.

        Returns
        -------
        self.model : `object`
            The object encapsulating the trained model.
        """
        model = kwargs.get('model', None)
        if model is None or model == 'None':  # pragma: no cover
            self.model = None
            return self.model
        if isinstance(model, str):
            self.model = self.set_data('model', data=None, path=model)
            self.config['model'] = model
            return self.model
        if isinstance(model, ModelHandle):
            if model.has_path:
                self.config['model'] = model.path
        self.model = self.set_data('model', model)
        return self.model

    def summarize(self, input_data, spec_data):
        """The main run method for the summarization, should be implemented
        in the specific subclass.

        This will attach the input_data to this `SZandPhottoNZSummarizer`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Finally, this will return a QPHandle providing access to that output data.

        Parameters
        ----------
        input_data : `qp.Ensemble`
            Per-galaxy p(z), and any ancilary data associated with it

        Returns
        -------
        output: `qp.Ensemble`
            Ensemble with n(z), and any ancilary data
        """
        self.set_data('input', input_data)
        self.set_data('spec_input', spec_data)
        self.run()
        self.finalize()
        return self.get_handle('output')


class PzInformer(RailStage):  #pragma: no cover
    """The base class for informing models used to summarize photo-z posterior estimates
    from ensembles of p(z) distributions.

    PzSummarizers can use a generic "model", the details of which depends on the sub-class.
    Some summaer will have associated PzInformer classes, which can be used to inform
    those models.

    (Note, "Inform" is more generic than "Train" as it also applies to algorithms that
    are template-based rather than machine learning-based.)

    PzInformer will produce as output a generic "model", the details of which depends on the sub-class.

    They take as "input" a qp.Ensemble of per-galaxy p(z) data, which is used to "inform" the model.
    """

    name = 'Informer'
    config_options = RailStage.config_options.copy()
    inputs = [('input', QPHandle)]
    outputs = [('model', ModelHandle)]

    def __init__(self, args, comm=None):
        """Initialize Informer that can inform models for redshift estimation """
        RailStage.__init__(self, args, comm=comm)
        self.model = None

    def inform(self, training_data):
        """The main interface method for Informers

        This will attach the input_data to this `Informer`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the model that it creates to this Estimator
        by using `self.add_data('model', model)`.

        Finally, this will return a ModelHandle providing access to the trained model.

        Parameters
        ----------
        input_data :  `qp.Ensemble`
            Per-galaxy p(z), and any ancilary data associated with it

        Returns
        -------
        model : ModelHandle
            Handle providing access to trained model
        """
        self.set_data('input', training_data)
        self.run()
        self.finalize()
        return self.get_handle('model')



import numpy as np
from .base import MetricEvaluator
from rail.evaluation.utils import stat_and_pval


class CDELoss(MetricEvaluator):
    """ Conditional density loss """
    def __init__(self, qp_ens, zgrid, ztrue):
        """Class constructor"""
        super().__init__(qp_ens)

        self._pdfs = qp_ens.pdf(zgrid)
        self._xvals = zgrid
        self._ztrue = ztrue
        self._npdf = qp_ens.npdf

    def evaluate(self):
        """Evaluate the estimated conditional density loss described in
        Izbicki & Lee 2017 (arXiv:1704.08095).

        Notes
        -----
        """

        # Calculate first term E[\int f*(z | X)^2 dz]
        term1 = np.mean(np.trapz(self._pdfs ** 2, x=self._xvals))
        # z bin closest to ztrue
        nns = [np.argmin(np.abs(self._xvals - z)) for z in self._ztrue]
        # Calculate second term E[f*(Z | X)]
        term2 = np.mean(self._pdfs[range(self._npdf), nns])
        cdeloss = term1 - 2 * term2
        return stat_and_pval(cdeloss, np.nan)



"""
Abstract base classes defining redshift estimations Informers and Estimators
"""

from rail.core.data import TableHandle, QPHandle, ModelHandle
from rail.core.stage import RailStage
import gc

class CatEstimator(RailStage):
    """The base class for making photo-z posterior estimates from catalog-like inputs
    (i.e., tables with fluxes in photometric bands among the set of columns)

    Estimators use a generic "model", the details of which depends on the sub-class.

    They take as "input" tabular data, apply the photo-z estimation and
    provide as "output" a QPEnsemble, with per-object p(z).

    """

    name = 'CatEstimator'
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=10000, hdf5_groupname=str)
    inputs = [('model', ModelHandle),
              ('input', TableHandle)]
    outputs = [('output', QPHandle)]

    def __init__(self, args, comm=None):
        """Initialize Estimator"""
        RailStage.__init__(self, args, comm=comm)
        self._output_handle = None
        self.model = None
        if not isinstance(args, dict):  #pragma: no cover
            args = vars(args)
        self.open_model(**args)

    def open_model(self, **kwargs):
        """Load the mode and/or attach it to this Estimator

        Parameters
        ----------
        model : `object`, `str` or `ModelHandle`
            Either an object with a trained model,
            a path pointing to a file that can be read to obtain the trained model,
            or a `ModelHandle` providing access to the trained model.

        Returns
        -------
        self.model : `object`
            The object encapsulating the trained model.
        """
        model = kwargs.get('model', None)
        if model is None or model == 'None':
            self.model = None
            return self.model
        if isinstance(model, str):
            self.model = self.set_data('model', data=None, path=model)
            self.config['model'] = model
            return self.model
        if isinstance(model, ModelHandle):
            if model.has_path:
                self.config['model'] = model.path
        self.model = self.set_data('model', model)
        return self.model

    def estimate(self, input_data):
        """The main interface method for the photo-z estimation

        This will attach the input_data to this `Estimator`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Finally, this will return a QPHandle providing access to that output data.

        Parameters
        ----------
        input_data : `dict` or `ModelHandle`
            Either a dictionary of all input data or a `ModelHandle` providing access to the same

        Returns
        -------
        output: `QPHandle`
            Handle providing access to QP ensemble with output data
        """
        self.set_data('input', input_data)
        self.run()
        self.finalize()
        return self.get_handle('output')

    def run(self):
        iterator = self.input_iterator('input')
        first = True
        self._initialize_run()
        self._output_handle = None
        for s, e, test_data in iterator:
            print(f"Process {self.rank} running estimator on chunk {s} - {e}")
            self._process_chunk(s, e, test_data, first)
            first = False
            # Running garbage collection manually seems to be needed
            # to avoid memory growth for some estimators
            gc.collect()
        self._finalize_run()

    def _initialize_run(self):
        self._output_handle = None

    def _finalize_run(self):
        self._output_handle.finalize_write()

    def _process_chunk(self, start, end, data, first):
        raise NotImplementedError(f"{self.name}._process_chunk is not implemented")  #pragma: no cover

    def _do_chunk_output(self, qp_dstn, start, end, first):
        if first:
            self._output_handle = self.add_handle('output', data = qp_dstn)
            self._output_handle.initialize_write(self._input_length, communicator = self.comm)
        self._output_handle.set_data(qp_dstn, partial=True)
        self._output_handle.write_chunk(start, end)



class CatInformer(RailStage):
    """The base class for informing models used to make photo-z posterior estimates
    from catalog-like inputs (i.e., tables with fluxes in photometric bands among
    the set of columns).

    Estimators use a generic "model", the details of which depends on the sub-class.
    Most estimators will have associated Informer classes, which can be used to inform
    those models.

    (Note, "Inform" is more generic than "Train" as it also applies to algorithms that
    are template-based rather than machine learning-based.)

    Informer will produce as output a generic "model", the details of which depends on the sub-class.

    They take as "input" catalog-like tabular data, which is used to "inform" the model.
    """

    name = 'Informer'
    config_options = RailStage.config_options.copy()
    config_options.update(hdf5_groupname=str, save_train=True)
    inputs = [('input', TableHandle)]
    outputs = [('model', ModelHandle)]

    def __init__(self, args, comm=None):
        """Initialize Informer that can inform models for redshift estimation """
        RailStage.__init__(self, args, comm=comm)
        self.model = None

    def inform(self, training_data):
        """The main interface method for Informers

        This will attach the input_data to this `Informer`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the model that it creates to this Estimator
        by using `self.add_data('model', model)`.

        Finally, this will return a ModelHandle providing access to the trained model.

        Parameters
        ----------
        input_data : `dict` or `TableHandle`
            dictionary of all input data, or a `TableHandle` providing access to it

        Returns
        -------
        model : ModelHandle
            Handle providing access to trained model
        """
        self.set_data('input', training_data)
        self.run()
        self.finalize()
        return self.get_handle('model')



import glob
from rail.cli import scripts
from rail.cli.options import GitMode

def test_render_nb():
    nb_files = glob.glob('../examples/*_examples/*.ipynb')
    scripts.render_nb('docs', False, True, nb_files)
    scripts.render_nb('docs', True, True, nb_files)


def test_clone_source():
    scripts.clone_source('..', GitMode.ssh, True, 'rail_packages.yml')
    scripts.clone_source('..', GitMode.https, True, 'rail_packages.yml')
    scripts.clone_source('..', GitMode.cli, True, 'rail_packages.yml')


def test_install():
    scripts.install('..', False, True, 'rail_packages.yml')
    scripts.install('..', True, True, 'rail_packages.yml')


def test_info():
    scripts.info(print_all=True)
